"""Tests for parrot.bots.flow.cel_evaluator — CEL Predicate Evaluator.

TASK-012: CEL compilation, evaluation, Pydantic coercion, fail-safe.
"""
import pytest
from pydantic import BaseModel

from parrot.bots.flow.cel_evaluator import CELPredicateEvaluator


class DecisionResult(BaseModel):
    final_decision: str
    confidence: float


# ---------------------------------------------------------------------------
# Compilation Tests
# ---------------------------------------------------------------------------

class TestCELCompilation:
    def test_valid_expression_compiles(self):
        evaluator = CELPredicateEvaluator('result.value == "test"')
        assert evaluator.expression == 'result.value == "test"'

    def test_invalid_expression_raises(self):
        with pytest.raises(ValueError, match="Invalid CEL expression"):
            CELPredicateEvaluator("result..value")

    def test_repr(self):
        evaluator = CELPredicateEvaluator('result == "x"')
        assert "CELPredicateEvaluator" in repr(evaluator)


# ---------------------------------------------------------------------------
# Simple Evaluation Tests
# ---------------------------------------------------------------------------

class TestCELEvaluation:
    def test_simple_equality_true(self):
        evaluator = CELPredicateEvaluator('result.decision == "pizza"')
        assert evaluator({"decision": "pizza"}) is True

    def test_simple_equality_false(self):
        evaluator = CELPredicateEvaluator('result.decision == "pizza"')
        assert evaluator({"decision": "sushi"}) is False

    def test_numeric_gt(self):
        evaluator = CELPredicateEvaluator("result.confidence > 0.8")
        assert evaluator({"confidence": 0.9}) is True
        assert evaluator({"confidence": 0.7}) is False

    def test_numeric_lt(self):
        evaluator = CELPredicateEvaluator("result.score < 50")
        assert evaluator({"score": 30}) is True
        assert evaluator({"score": 80}) is False

    def test_boolean_and(self):
        evaluator = CELPredicateEvaluator(
            "result.approved && result.confidence > 0.5"
        )
        assert evaluator({"approved": True, "confidence": 0.8}) is True
        assert evaluator({"approved": True, "confidence": 0.3}) is False
        assert evaluator({"approved": False, "confidence": 0.9}) is False

    def test_boolean_or(self):
        evaluator = CELPredicateEvaluator(
            'result.status == "ok" || result.status == "accepted"'
        )
        assert evaluator({"status": "ok"}) is True
        assert evaluator({"status": "accepted"}) is True
        assert evaluator({"status": "rejected"}) is False

    def test_negation(self):
        evaluator = CELPredicateEvaluator('!(result.failed)')
        assert evaluator({"failed": False}) is True
        assert evaluator({"failed": True}) is False

    def test_string_equality(self):
        evaluator = CELPredicateEvaluator('result == "hello"')
        assert evaluator("hello") is True
        assert evaluator("world") is False

    def test_in_operator_list(self):
        evaluator = CELPredicateEvaluator(
            'result.category in ["A", "B", "C"]'
        )
        assert evaluator({"category": "A"}) is True
        assert evaluator({"category": "Z"}) is False


# ---------------------------------------------------------------------------
# Context & Error Tests
# ---------------------------------------------------------------------------

class TestCELContextAccess:
    def test_ctx_access(self):
        evaluator = CELPredicateEvaluator("ctx.retries < 3")
        assert evaluator({}, retries=2) is True
        assert evaluator({}, retries=5) is False

    def test_error_variable_with_exception(self):
        evaluator = CELPredicateEvaluator('error != ""')
        assert evaluator({}, error=Exception("failed")) is True

    def test_error_variable_no_error(self):
        evaluator = CELPredicateEvaluator('error != ""')
        assert evaluator({}, error=None) is False


# ---------------------------------------------------------------------------
# Pydantic Coercion Tests
# ---------------------------------------------------------------------------

class TestPydanticCoercion:
    def test_pydantic_model_coerced(self):
        evaluator = CELPredicateEvaluator('result.final_decision == "pizza"')
        result = DecisionResult(final_decision="pizza", confidence=0.95)
        assert evaluator(result) is True

    def test_pydantic_nested_access(self):
        evaluator = CELPredicateEvaluator("result.confidence > 0.9")
        result = DecisionResult(final_decision="pizza", confidence=0.95)
        assert evaluator(result) is True

    def test_pydantic_model_false(self):
        evaluator = CELPredicateEvaluator('result.final_decision == "sushi"')
        result = DecisionResult(final_decision="pizza", confidence=0.5)
        assert evaluator(result) is False


# ---------------------------------------------------------------------------
# Error Handling (fail-safe)
# ---------------------------------------------------------------------------

class TestErrorHandling:
    def test_missing_field_returns_false(self):
        evaluator = CELPredicateEvaluator('result.missing_field == "value"')
        assert evaluator({"other": "data"}) is False

    def test_type_mismatch_returns_false(self):
        evaluator = CELPredicateEvaluator("result.value > 10")
        assert evaluator({"value": "not a number"}) is False

    def test_none_result(self):
        evaluator = CELPredicateEvaluator('result == "test"')
        assert evaluator(None) is False


# ---------------------------------------------------------------------------
# Import Tests
# ---------------------------------------------------------------------------

class TestImports:
    def test_import_from_package(self):
        from parrot.bots.flow import CELPredicateEvaluator as CEL

        assert CEL is CELPredicateEvaluator



import asyncio
import sys
import importlib.util
from importlib import import_module
from pathlib import Path
from typing import Optional
import yaml
import click
from navconfig.logging import logging
from .server import MCPServer, MCPServerConfig
from ..tools.abstract import AbstractTool
from ..tools.toolkit import AbstractToolkit
from ..services.mcp.server import ParrotMCPServer, TransportConfig


@click.group(invoke_without_command=True)
@click.option('--config', type=click.Path(exists=True), help='Path to YAML configuration file')
@click.pass_context
def mcp(ctx, config):
    """MCP server commands."""
    if ctx.invoked_subcommand is None:
        if config:
            # Run server from config
            from .wrapper import load_server_from_config
            try:
                server = load_server_from_config(config)
                # SimpleMCPServer.run() is blocking and handles the loop internally for http/sse
                # If we need async start for stdio/etc and simple.py run() does it, we assume it works.
                server.run()
            except Exception as e:
                click.echo(f"Error starting server: {e}", err=True)
                sys.exit(1)
        else:
            click.echo(ctx.get_help())


@mcp.command()
@click.argument('config_file', type=click.Path(exists=True))
@click.option(
    '--transport', type=click.Choice(['stdio', 'unix', 'http']), default=None,
        help='Override transport from config')
@click.option(
    '--socket', type=str, default=None,
            help='Unix socket path (for unix transport)')
@click.option(
    '--port', type=int, default=None,
            help='Port (for http transport)')
@click.option(
    '--log-level', type=str, default='INFO',
            help='Logging level')
def serve(
    config_file: str, transport: Optional[str], socket: Optional[str],
        port: Optional[int], log_level: str):
    """
    Start an MCP server from a Python config file or YAML.

    Examples:

        # Python config file
        parrot mcp serve workday_server.py --transport unix --socket /tmp/workday.sock

        # YAML config file
        parrot mcp serve mcp_config.yaml

    Python config file should define 'mcp' variable:

        # workday_server.py
        from parrot.services import ParrotMCPServer
        from parrot.toolkits.workday import WorkdayToolkit

        mcp = ParrotMCPServer(
            name="workday-mcp",
            tools=WorkdayToolkit(redis_url="redis://localhost:6379/4")
        )
    """
    config_path = Path(config_file)

    if config_path.suffix in {'.yaml', '.yml'}:
        mcp_server = _load_from_yaml(config_path)
    elif config_path.suffix == '.py':
        mcp_server = _load_from_python(config_path)
    else:
        click.echo(f"Error: Unsupported config file type: {config_path.suffix}", err=True)
        sys.exit(1)

    # Override settings from CLI
    if transport:
        # Need to update transport config
        mcp_server.transport_configs = {
            transport: _create_transport_config(transport, socket, port)
        }

    # Set log level
    logging.getLogger().setLevel(log_level)

    # Run the server
    asyncio.run(_run_standalone_server(mcp_server))


def _load_from_python(config_path: Path) -> 'ParrotMCPServer':
    """Load ParrotMCPServer from Python file."""


    spec = importlib.util.spec_from_file_location("mcp_config", config_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if not hasattr(module, 'mcp'):
        raise ValueError(
            f"Config file {config_path} must define 'mcp' variable "
            "containing a ParrotMCPServer instance"
        )

    return module.mcp


def _load_from_yaml(config_path: Path) -> ParrotMCPServer:
    """Load ParrotMCPServer from YAML file."""
    with open(config_path) as f:
        config = yaml.safe_load(f)

    # Parse tools from YAML
    tools_config = {
        tool_entry['class']: tool_entry['module']
        for tool_entry in config.get('tools', [])
    }

    # Parse transports
    transports = config.get('transport', 'stdio')

    return ParrotMCPServer(
        name=config.get('name', 'ai-parrot-mcp'),
        description=config.get('description', 'AI-Parrot MCP Server'),
        transports=transports,
        tools=tools_config,
        **config.get('server_config', {})
    )


def _create_transport_config(transport: str, socket: Optional[str], port: Optional[int]):
    """Create TransportConfig from CLI args."""
    return TransportConfig(
        transport=transport,
        host="127.0.0.1" if transport == "http" else None,
        port=port if transport == "http" else None,
    )


async def _run_standalone_server(mcp_server: ParrotMCPServer):
    """Run MCP server in standalone mode (no aiohttp app)."""
    logger = logging.getLogger("parrot.mcp.serve")

    # Load tools
    tools = await mcp_server._load_configured_tools()
    if not tools:
        logger.error("No tools configured")
        sys.exit(1)

    logger.info(f"Loaded {len(tools)} tools")

    # Get transport config (should be single transport in CLI mode)
    if len(mcp_server.transport_configs) != 1:
        logger.error("CLI mode requires exactly one transport")
        sys.exit(1)

    transport_key, transport_config = list(mcp_server.transport_configs.items())[0]

    # Create MCP server
    config = MCPServerConfig(
        name=mcp_server.name,
        description=mcp_server.description,
        transport=transport_config.transport,
        host=transport_config.host,
        port=transport_config.port,
        socket_path=transport_config.socket_path if hasattr(transport_config, 'socket_path') else None,
        log_level=mcp_server.log_level,
    )

    server = MCPServer(config)
    server.register_tools(tools)

    # Start and run
    try:
        logger.info(f"Starting MCP server in {transport_config.transport} mode...")
        await server.start()
    except KeyboardInterrupt:
        logger.info("Interrupted by user")
    finally:
        await server.stop()



from pathlib import Path

from parrot.integrations.parser import ParsedResponse
from parrot.integrations.slack.wrapper import SlackAgentWrapper
from parrot.integrations.models import IntegrationBotConfig


def test_build_blocks_supports_markdown_code_table_and_images():
    parsed = ParsedResponse(
        text="*hello*",
        code="print('x')",
        code_language="python",
        table_markdown="|A|\n|---|\n|1|",
        images=[Path("https://example.com/img.png"), Path("/tmp/local.png")],
    )

    blocks = SlackAgentWrapper._build_blocks(parsed)

    assert any(b.get("type") == "section" and b["text"]["type"] == "mrkdwn" for b in blocks)
    assert any(b.get("type") == "image" for b in blocks)
    assert any(b.get("type") == "context" for b in blocks)


def test_integration_config_parses_slack_kind():
    cfg = IntegrationBotConfig.from_dict(
        {
            "agents": {
                "sales": {
                    "kind": "slack",
                    "chatbot_id": "sales_bot",
                    "bot_token": "xoxb-123",
                }
            }
        }
    )

    assert "sales" in cfg.agents
    assert cfg.validate() == []



"""Configuration for WhatsApp Bridge integration."""
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Any

from navconfig import config


@dataclass
class WhatsAppBridgeConfig:
    """Configuration for WhatsApp Bridge wrapper (whatsmeow-based).

    Attributes:
        name: Wrapper name (used for logging and route generation).
        chatbot_id: Agent name in BotManager / agent registry.
        bridge_url: URL of the Go whatsmeow bridge.
        webhook_path: Path to register for incoming message callbacks.
        welcome_message: Greeting sent on first interaction.
        system_prompt_override: Override agent's default system prompt.
        allowed_numbers: Phone allowlist (digits only, no +). Empty = all.
        commands: Custom slash-command map.
        max_message_length: Max chars before splitting.
    """

    name: str
    chatbot_id: str
    bridge_url: str = "http://localhost:8765"
    webhook_path: Optional[str] = None
    welcome_message: Optional[str] = None
    system_prompt_override: Optional[str] = None
    commands: Dict[str, str] = field(default_factory=dict)
    allowed_numbers: Optional[List[str]] = None
    max_message_length: int = 4096

    def __post_init__(self):
        """Resolve bridge_url from environment if not set."""
        if not self.bridge_url:
            self.bridge_url = config.get(
                "WHATSAPP_BRIDGE_URL", "http://localhost:8765"
            )

    @classmethod
    def from_dict(cls, name: str, data: Dict[str, Any]) -> "WhatsAppBridgeConfig":
        """Create config from dictionary (YAML parsed data)."""
        return cls(
            name=name,
            chatbot_id=data.get("chatbot_id", name),
            bridge_url=data.get("bridge_url", "http://localhost:8765"),
            webhook_path=data.get("webhook_path"),
            welcome_message=data.get("welcome_message"),
            system_prompt_override=data.get("system_prompt_override"),
            commands=data.get("commands", {}),
            allowed_numbers=data.get("allowed_numbers"),
            max_message_length=data.get("max_message_length", 4096),
        )



from typing import Any
import json
try:
    from rich.console import Console, Group
    from rich.panel import Panel
    from rich.markdown import Markdown
    from rich.table import Table
    from rich.layout import Layout
    from rich.syntax import Syntax
    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False
from .abstract import AbstractAppGenerator


class TerminalGenerator(AbstractAppGenerator):
    """Generates a Rich Console Dashboard."""

    def generate(self) -> Any:
        if not RICH_AVAILABLE:
            return f"Query: {self.payload['input']}\n\n{self.payload['explanation']}"

        # 1. Header
        header = Panel(
            f"[bold blue]Query:[/bold blue] {self.payload['input']}",
            title="🤖 AI Agent Analysis",
            border_style="green"
        )

        # 2. Explanation (Markdown)
        explanation = Panel(
            Markdown(self.payload["explanation"]),
            title="📝 Analysis",
            border_style="blue"
        )

        # 3. Data Table
        df = self.payload["data"]
        if not df.empty:
            table = Table(show_header=True, header_style="bold magenta")
            # Limit to first 5 columns and 10 rows for terminal readability
            show_cols = df.columns[:5]
            for col in show_cols:
                table.add_column(str(col))

            for _, row in df.head(10).iterrows():
                table.add_row(*[str(item) for item in row])

            data_panel = Panel(table, title=f"🔢 Data Preview ({len(df)} rows)", border_style="yellow")
        else:
            data_panel = Panel("No structured data returned.", title="🔢 Data", border_style="yellow")

        # 4. Code View
        if code_content := self.payload["code"]:
            if isinstance(code_content, (dict, list)):
                syntax = Syntax(json.dumps(code_content, indent=2), "json", theme="monokai")
            else:
                syntax = Syntax(code_content, "python", theme="monokai", line_numbers=True)

            code_panel = Panel(syntax, title="📊 Generated Code", border_style="cyan")
            return Group(header, explanation, data_panel, code_panel)

        return Group(header, explanation, data_panel)



import base64
import time
from typing import Any, Dict, Optional
import aiohttp
from navconfig.logging import logging

class ZoomUsInterface:
    """
    Interface for interacting with Zoom.us API via Server-to-Server OAuth.
    """
    BASE_URL = "https://api.zoom.us/v2"
    AUTH_URL = "https://zoom.us/oauth/token"

    def __init__(
        self,
        account_id: str,
        client_id: str,
        client_secret: str,
        logger: Optional[logging.Logger] = None
    ):
        self.account_id = account_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.logger = logger or logging.getLogger("ZoomUsInterface")
        self._token: Optional[str] = None
        self._token_expires_at: float = 0
        self._session: Optional[aiohttp.ClientSession] = None

    async def connect(self):
        """Initialize the session."""
        if not self._session:
            self._session = aiohttp.ClientSession()

    async def close(self):
        """Close the session."""
        if self._session:
            await self._session.close()
            self._session = None

    async def _get_access_token(self) -> str:
        """
        Get or refresh the Server-to-Server OAuth access token.
        """
        if self._token and time.time() < self._token_expires_at - 60:
            return self._token

        if not self._session:
            await self.connect()

        auth_str = f"{self.client_id}:{self.client_secret}"
        b64_auth = base64.b64encode(auth_str.encode()).decode()

        params = {
            "grant_type": "account_credentials",
            "account_id": self.account_id,
        }
        headers = {
            "Authorization": f"Basic {b64_auth}",
            "Content-Type": "application/x-www-form-urlencoded"
        }

        self.logger.debug("Requesting new Zoom access token...")
        async with self._session.post(self.AUTH_URL, params=params, headers=headers) as resp:
            if resp.status != 200:
                text = await resp.text()
                self.logger.error(f"Failed to get Zoom token: {resp.status} - {text}")
                raise Exception(f"Failed to get Zoom token: {resp.status}")

            data = await resp.json()
            self._token = data["access_token"]
            expires_in = data.get("expires_in", 3600)
            self._token_expires_at = time.time() + expires_in
            self.logger.info("Successfully obtained Zoom access token.")
            return self._token

    async def request(self, method: str, endpoint: str, **kwargs) -> Dict[str, Any]:
        """
        Make an authenticated request to the Zoom API.
        """
        token = await self._get_access_token()
        headers = kwargs.get("headers", {})
        headers["Authorization"] = f"Bearer {token}"
        headers["Content-Type"] = "application/json"
        kwargs["headers"] = headers

        url = f"{self.BASE_URL}{endpoint}"

        if not self._session:
            await self.connect()

        async with self._session.request(method, url, **kwargs) as resp:
            if resp.status == 401:
                # Token might be expired, retry once
                self.logger.warning("Zoom API 401, retrying with new token...")
                self._token = None
                token = await self._get_access_token()
                headers["Authorization"] = f"Bearer {token}"
                async with self._session.request(method, url, **kwargs) as resp_retry:
                    return await self._handle_response(resp_retry)
            
            return await self._handle_response(resp)

    async def _handle_response(self, response: aiohttp.ClientResponse) -> Dict[str, Any]:
        """Handle response processing."""
        if response.status >= 400:
            text = await response.text()
            self.logger.error(f"Zoom API Error {response.status}: {text}")
            raise Exception(f"Zoom API Error {response.status}: {text}")
        
        try:
            return await response.json()
        except:
            return {}

    async def get_account_settings(self, option: str = None, **kwargs) -> Dict[str, Any]:
        """
        Get Account Settings.
        https://developers.zoom.us/docs/api/rest/reference/phone/methods/#operation/getAccountSettings
        """
        params = kwargs
        if option:
            params['option'] = option
            
        return await self.request("GET", "/phone/account_settings", params=params)



"""HookManager — registry and lifecycle coordinator for all hooks."""
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

from navconfig.logging import logging

from .base import BaseHook

if TYPE_CHECKING:
    from parrot.core.events.evb import EventBus


class HookManager:
    """Manages registration, startup, and shutdown of all external hooks.

    The manager injects a callback into each hook so that fired events
    flow into the orchestrator's execution pipeline.

    Optionally, an :class:`EventBus` can be attached via
    :meth:`set_event_bus` to enable distributed dual-emit: every hook
    event is forwarded to both the direct callback *and* the bus on
    channel ``hooks.<hook_type>.<event_type>``.
    """

    def __init__(self) -> None:
        self._hooks: Dict[str, BaseHook] = {}
        self._callback: Optional[Callable] = None
        self._event_bus: Optional["EventBus"] = None
        self.logger = logging.getLogger("parrot.hooks.manager")

    def set_event_callback(self, callback) -> None:
        """Set the async callback that all hooks will invoke on events.

        Typically ``AutonomousOrchestrator._handle_hook_event``.
        """
        self._callback = callback
        dispatch = self._build_dispatch()
        for hook in self._hooks.values():
            hook.set_callback(dispatch)

    def set_event_bus(self, bus: "EventBus") -> None:
        """Attach an :class:`EventBus` for distributed event publishing.

        When set, every hook event is emitted to both the registered
        callback *and* the bus on channel
        ``hooks.<hook_type>.<event_type>``.  If no bus is set, the
        existing callback-only behaviour is preserved unchanged.

        Args:
            bus: An :class:`~parrot.core.events.evb.EventBus` instance.
        """
        self._event_bus = bus
        dispatch = self._build_dispatch()
        for hook in self._hooks.values():
            hook.set_callback(dispatch)
        self.logger.info("HookManager: EventBus attached — dual-emit enabled")

    def _build_dispatch(self) -> Optional[Callable]:
        """Return the effective per-hook callback.

        * No bus → return the raw user callback unchanged.
        * Bus set → return an ``_dual_emit`` wrapper that calls the
          callback *and* emits to the bus.  Either the callback or the
          bus may be absent individually without raising.

        Closure strategy
        ----------------
        ``bus`` is captured at build time (it is invariant once set).
        The user callback is read from ``self._callback`` **at dispatch
        time** — not captured — so hooks built before
        ``set_event_callback()`` is called still see the correct
        callback without needing re-injection.  This eliminates the
        ordering-hazard window where events fired between
        ``set_event_bus()`` and ``set_event_callback()`` would silently
        drop the callback.

        Both sync and async callbacks are supported via
        ``asyncio.iscoroutinefunction()`` inspection.
        """
        bus = self._event_bus

        if bus is None:
            return self._callback

        async def _dual_emit(event) -> None:
            # Read callback at call time — avoids stale-closure issue when
            # set_event_callback() is called after this dispatch was built.
            cb = self._callback
            if cb is not None:
                if asyncio.iscoroutinefunction(cb):
                    await cb(event)
                else:
                    cb(event)
            try:
                await bus.emit(
                    f"hooks.{event.hook_type.value}.{event.event_type}",
                    event.model_dump(),
                )
            except Exception as exc:
                self.logger.warning(
                    "HookManager: EventBus emit failed for %s.%s: %s",
                    event.hook_type.value,
                    event.event_type,
                    exc,
                )

        return _dual_emit

    def register(self, hook: BaseHook) -> str:
        """Register a hook and return its hook_id.

        If a callback is already set, it is injected into the hook
        immediately so it is ready before ``start_all()`` is called.
        """
        if hook.hook_id in self._hooks:
            self.logger.warning(
                f"Hook '{hook.hook_id}' already registered, replacing"
            )
        self._hooks[hook.hook_id] = hook
        dispatch = self._build_dispatch()
        if dispatch is not None:
            hook.set_callback(dispatch)
        self.logger.info(f"Registered hook: {hook!r}")
        return hook.hook_id

    def unregister(self, hook_id: str) -> Optional[BaseHook]:
        """Unregister a hook by ID. Returns the removed hook or None."""
        hook = self._hooks.pop(hook_id, None)
        if hook:
            self.logger.info(f"Unregistered hook: {hook!r}")
        return hook

    def get_hook(self, hook_id: str) -> Optional[BaseHook]:
        """Retrieve a registered hook by ID."""
        return self._hooks.get(hook_id)

    async def start_all(self) -> None:
        """Start all enabled hooks."""
        started = 0
        for hook in self._hooks.values():
            if not hook.enabled:
                self.logger.debug(f"Skipping disabled hook: {hook.name}")
                continue
            try:
                await hook.start()
                started += 1
                self.logger.info(f"Started hook: {hook!r}")
            except Exception as exc:
                self.logger.error(
                    f"Failed to start hook '{hook.name}': {exc}",
                    exc_info=True,
                )
        self.logger.info(
            f"HookManager started {started}/{len(self._hooks)} hooks"
        )

    async def stop_all(self) -> None:
        """Stop all running hooks."""
        stopped = 0
        for hook in self._hooks.values():
            if not hook.enabled:
                continue
            try:
                await hook.stop()
                stopped += 1
                self.logger.debug(f"Stopped hook: {hook!r}")
            except Exception as exc:
                self.logger.error(
                    f"Failed to stop hook '{hook.name}': {exc}",
                    exc_info=True,
                )
        self.logger.info(
            f"HookManager stopped {stopped} hooks"
        )

    def setup_routes(self, app: Any) -> None:
        """Delegate route setup to HTTP-based hooks."""
        for hook in self._hooks.values():
            if hook.enabled:
                hook.setup_routes(app)

    @property
    def hooks(self) -> List[BaseHook]:
        """List all registered hooks."""
        return list(self._hooks.values())

    @property
    def stats(self) -> Dict[str, Any]:
        """Return summary statistics."""
        return {
            "total": len(self._hooks),
            "enabled": sum(1 for h in self._hooks.values() if h.enabled),
            "by_type": self._count_by_type(),
        }

    def _count_by_type(self) -> Dict[str, int]:
        counts: Dict[str, int] = {}
        for hook in self._hooks.values():
            key = hook.hook_type.value
            counts[key] = counts.get(key, 0) + 1
        return counts



"""
FAISSStore: In-memory Vector Store implementation using FAISS.

Provides high-performance vector similarity search with:
- In-memory vector storage with FAISS indexes
- Multiple distance metrics (Cosine, L2, Inner Product)
- CPU-only execution (GPU support removed)
- MMR (Maximal Marginal Relevance) search
- Metadata filtering
- Collection management
- Async context manager support
"""
from typing import Any, Dict, List, Optional, Union, Callable
import uuid
import pickle
from pathlib import Path
import numpy as np
from navconfig.logging import logging
from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances

try:
    import faiss
    FAISS_AVAILABLE = True
except ImportError:
    FAISS_AVAILABLE = False
    faiss = None

from .abstract import AbstractStore
from .models import Document, SearchResult, DistanceStrategy


class FAISSStore(AbstractStore):
    """
    An in-memory FAISS vector store implementation, completely independent of Langchain.

    This store provides high-performance vector similarity search using FAISS indexes
    with support for multiple distance metrics and metadata filtering.

    Features:
    - Multiple FAISS index types (Flat, IVF, HNSW)
    - CPU-only execution
    - Cosine, L2, and Inner Product distance metrics
    - MMR (Maximal Marginal Relevance) search
    - Metadata filtering
    - Persistent storage via save/load
    """

    def __init__(
        self,
        collection_name: str = "default",
        id_column: str = 'id',
        embedding_column: str = 'embedding',
        document_column: str = 'document',
        text_column: str = 'text',
        metadata_column: str = 'metadata',
        embedding_model: Union[dict, str] = "sentence-transformers/all-mpnet-base-v2",
        embedding: Optional[Callable] = None,
        distance_strategy: DistanceStrategy = DistanceStrategy.COSINE,
        index_type: str = "Flat",  # Options: "Flat", "IVF", "HNSW"
        nlist: int = 100,  # For IVF indexes
        nprobe: int = 10,  # For IVF search
        m: int = 32,  # For HNSW
        ef_construction: int = 40,  # For HNSW
        ef_search: int = 16,  # For HNSW
        **kwargs
    ):
        """
        Initialize FAISSStore with the specified parameters.

        Args:
            collection_name: Name of the collection/index
            id_column: Name of the ID column
            embedding_column: Name of the embedding column
            document_column: Name of the document content column
            text_column: Name of the text column
            metadata_column: Name of the metadata column
            embedding_model: Embedding model configuration (dict or string)
            embedding: Custom embedding function
            distance_strategy: Distance metric to use (COSINE, EUCLIDEAN_DISTANCE, etc.)
            index_type: Type of FAISS index ("Flat", "IVF", "HNSW")
            nlist: Number of clusters for IVF indexes
            nprobe: Number of clusters to probe for IVF search
            m: Number of connections per layer for HNSW
            ef_construction: Size of dynamic candidate list for HNSW construction
            ef_search: Size of dynamic candidate list for HNSW search
        """
        if not FAISS_AVAILABLE:
            raise ImportError(
                "FAISS is not installed. Please install it with: pip install faiss-cpu"
            )

        # Store configuration
        self.collection_name = collection_name
        self._id_column: str = id_column
        self._embedding_column: str = embedding_column
        self._document_column: str = document_column
        self._text_column: str = text_column
        self._metadata_column: str = metadata_column

        # FAISS configuration
        self.index_type = index_type
        self.nlist = nlist
        self.nprobe = nprobe
        self.m = m
        self.ef_construction = ef_construction
        self.ef_search = ef_search

        # Distance strategy - normalize to enum
        if isinstance(distance_strategy, str):
            # Convert string to DistanceStrategy enum
            try:
                self.distance_strategy = DistanceStrategy[distance_strategy.upper()]
            except KeyError:
                self.logger.warning(
                    f"Unknown distance strategy '{distance_strategy}', using COSINE"
                )
                self.distance_strategy = DistanceStrategy.COSINE
        elif isinstance(distance_strategy, DistanceStrategy):
            self.distance_strategy = distance_strategy
        else:
            # Default to COSINE if invalid type
            self.logger.warning(
                f"Invalid distance_strategy type: {type(distance_strategy)}, using COSINE"
            )
            self.distance_strategy = DistanceStrategy.COSINE

        # Initialize parent class
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )

        # Collections store: {collection_name: collection_data}
        self._collections: Dict[str, Dict[str, Any]] = {}

        # Initialize logger
        self.logger = logging.getLogger("FAISSStore")

        # Connection state
        self._connected: bool = False
        self._connection = None  # For compatibility with abstract interface

        # Initialize default collection
        if collection_name:
            self._initialize_collection(collection_name)

    def _initialize_collection(self, collection_name: str) -> None:
        """Initialize a new collection with empty data structures."""
        if collection_name not in self._collections:
            self._collections[collection_name] = {
                'index': None,  # FAISS index
                'documents': {},  # {id: document_content}
                'metadata': {},  # {id: metadata_dict}
                'embeddings': {},  # {id: embedding_vector}
                'id_to_idx': {},  # {id: faiss_index_position}
                'idx_to_id': {},  # {faiss_index_position: id}
                'dimension': None,
                'is_trained': False,
            }
            self.logger.info(f"Initialized collection: {collection_name}")

    def define_collection_table(
        self,
        collection_name: str,
        dimension: int = 384,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Define an in-memory collection table for saving vector + metadata information.

        This method is compatible with the PgVectorStore pattern but operates in-memory.

        Args:
            collection_name: Name of the collection
            dimension: Dimension of the embedding vectors
            **kwargs: Additional arguments (for compatibility)

        Returns:
            Dictionary representing the collection structure
        """
        if collection_name not in self._collections:
            self._initialize_collection(collection_name)

        collection = self._collections[collection_name]
        collection['dimension'] = dimension

        # Create FAISS index based on configuration
        index = self._create_faiss_index(dimension)
        collection['index'] = index

        self.logger.info(
            f"Defined collection table '{collection_name}' with dimension {dimension}"
        )

        return {
            'collection_name': collection_name,
            'dimension': dimension,
            'index_type': self.index_type,
            'distance_strategy': self.distance_strategy.value,
        }

    def _create_faiss_index(self, dimension: int) -> Any:
        """
        Create a FAISS index based on the configured index type and distance strategy.

        Args:
            dimension: Dimension of the embedding vectors

        Returns:
            FAISS index object
        """
        # Determine the metric type based on distance strategy
        if self.distance_strategy == DistanceStrategy.COSINE:
            # For cosine similarity, we'll normalize vectors and use inner product
            metric = faiss.METRIC_INNER_PRODUCT
        elif self.distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            metric = faiss.METRIC_L2
        elif self.distance_strategy in [DistanceStrategy.DOT_PRODUCT, DistanceStrategy.MAX_INNER_PRODUCT]:
            metric = faiss.METRIC_INNER_PRODUCT
        else:
            # Default to inner product
            metric = faiss.METRIC_INNER_PRODUCT

        # Create index based on type
        if self.index_type == "Flat":
            index = faiss.IndexFlatIP(dimension) if metric == faiss.METRIC_INNER_PRODUCT else faiss.IndexFlatL2(dimension)

        elif self.index_type == "IVF":
            # IVF (Inverted File Index) for faster search on large datasets
            quantizer = faiss.IndexFlatIP(dimension) if metric == faiss.METRIC_INNER_PRODUCT else faiss.IndexFlatL2(dimension)
            index = faiss.IndexIVFFlat(quantizer, dimension, self.nlist, metric)
            index.nprobe = self.nprobe

        elif self.index_type == "HNSW":
            # HNSW (Hierarchical Navigable Small World) for very fast search
            index = faiss.IndexHNSWFlat(dimension, self.m, metric)
            index.hnsw.efConstruction = self.ef_construction
            index.hnsw.efSearch = self.ef_search

        else:
            raise ValueError(f"Unsupported index type: {self.index_type}")

        self.logger.info(
            f"Created FAISS index: type={self.index_type}, "
            f"metric={metric}, dimension={dimension}, cpu_only=True"
        )

        return index

    async def connection(self) -> bool:
        """
        Establish connection (for compatibility with AbstractStore).

        Since FAISS is in-memory, this just marks the store as connected.

        Returns:
            True if successful
        """
        if not self._connected:
            self._connected = True
            self._connection = True  # Dummy connection for compatibility
            self.logger.info("FAISSStore connection established (in-memory)")
        return True

    async def disconnect(self) -> None:
        """
        Disconnect and cleanup resources.

        Clears all in-memory data.
        """
        if not self._connected:
            return
        # Clear indexes
        for _, collection in self._collections.items():
            if collection.get('index'):
                del collection['index']

        # Clear collections
        self._collections.clear()

        self._connected = False
        self._connection = None
        self.logger.info("FAISSStore disconnected and resources cleared")

    async def __aenter__(self):
        """Async context manager entry."""
        if not self._connected:
            await self.connection()
        self._context_depth += 1
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Async context manager exit."""
        # Free embedding resources
        if self._embed_:
            await self._free_resources()

        try:
            # Only disconnect if we're exiting the outermost context
            self._context_depth -= 1
            if self._context_depth <= 0:
                await self.disconnect()
                self._context_depth = 0
        except RuntimeError:
            pass

    async def prepare_embedding_table(
        self,
        collection: str = None,
        dimension: int = None,
        **kwargs
    ) -> bool:
        """
        Prepare the embedding table/collection for storing vectors.

        Args:
            collection: Collection name
            dimension: Embedding dimension
            **kwargs: Additional arguments

        Returns:
            True if successful
        """
        collection = collection or self.collection_name
        dimension = dimension or self.dimension

        if collection not in self._collections:
            self.define_collection_table(collection, dimension, **kwargs)
        else:
            # Update dimension if needed
            if dimension and self._collections[collection]['dimension'] != dimension:
                self._collections[collection]['dimension'] = dimension
                # Recreate index with new dimension
                self._collections[collection]['index'] = self._create_faiss_index(dimension)
                self.logger.info(f"Updated collection '{collection}' dimension to {dimension}")

        return True

    async def create_embedding_table(
        self,
        collection: str = None,
        dimension: int = None,
        **kwargs
    ) -> None:
        """
        Create an embedding table/collection (alias for prepare_embedding_table).

        Args:
            collection: Collection name
            dimension: Embedding dimension
            **kwargs: Additional arguments
        """
        await self.prepare_embedding_table(collection, dimension, **kwargs)

    async def create_collection(self, collection: str, **kwargs) -> None:
        """
        Create a new collection.

        Args:
            collection: Collection name
            **kwargs: Additional arguments (e.g., dimension)
        """
        dimension = kwargs.get('dimension', self.dimension)
        await self.create_embedding_table(collection, dimension, **kwargs)

    async def add_documents(
        self,
        documents: List[Document],
        collection: str = None,
        embedding_column: str = None,
        content_column: str = None,
        metadata_column: str = None,
        **kwargs
    ) -> None:
        """
        Add documents to the FAISS store.

        Args:
            documents: List of Document objects to add
            collection: Collection name (optional, uses default if not provided)
            embedding_column: Name of the embedding column (for compatibility)
            content_column: Name of the content column (for compatibility)
            metadata_column: Name of the metadata column (for compatibility)
            **kwargs: Additional arguments
        """
        if not self._connected:
            await self.connection()

        collection = collection or self.collection_name

        # Ensure collection exists
        if collection not in self._collections:
            self._initialize_collection(collection)

        collection_data = self._collections[collection]

        # Extract texts and metadata
        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]

        # Generate embeddings
        embeddings = await self._embed_.embed_documents(texts)

        # Convert to numpy array
        if isinstance(embeddings, list):
            embeddings = np.array(embeddings, dtype=np.float32)
        elif not isinstance(embeddings, np.ndarray):
            embeddings = np.array(embeddings, dtype=np.float32)

        # Ensure 2D array
        if embeddings.ndim == 1:
            embeddings = embeddings.reshape(1, -1)

        # Set dimension if not set
        if collection_data['dimension'] is None:
            collection_data['dimension'] = embeddings.shape[1]
            collection_data['index'] = self._create_faiss_index(embeddings.shape[1])

        # Normalize embeddings for cosine similarity
        if self.distance_strategy == DistanceStrategy.COSINE:
            faiss.normalize_L2(embeddings)

        # Train index if needed (for IVF)
        if self.index_type == "IVF" and not collection_data['is_trained']:
            if len(embeddings) >= self.nlist:
                collection_data['index'].train(embeddings)
                collection_data['is_trained'] = True
                self.logger.info(f"Trained IVF index for collection '{collection}'")
            else:
                self.logger.warning(
                    f"Not enough vectors to train IVF index "
                    f"(need {self.nlist}, got {len(embeddings)}). Using Flat index temporarily."
                )

        # Get current index size
        current_idx = collection_data['index'].ntotal

        # Add to FAISS index
        collection_data['index'].add(embeddings)

        # Store documents, metadata, and embeddings
        for i, (text, metadata, embedding) in enumerate(zip(texts, metadatas, embeddings)):
            doc_id = str(uuid.uuid4())
            idx = current_idx + i

            collection_data['documents'][doc_id] = text
            collection_data['metadata'][doc_id] = metadata or {}
            collection_data['embeddings'][doc_id] = embedding
            collection_data['id_to_idx'][doc_id] = idx
            collection_data['idx_to_id'][idx] = doc_id

        self.logger.info(
            f"✅ Successfully added {len(documents)} documents to collection '{collection}'"
        )

    def get_distance_strategy(
        self,
        query_embedding: np.ndarray,
        metric: str = None
    ) -> str:
        """
        Return the appropriate distance strategy based on the metric or configured strategy.

        Args:
            query_embedding: Query embedding vector (for compatibility)
            metric: Optional metric string ('COSINE', 'L2', 'IP', 'DOT')

        Returns:
            Distance strategy as string
        """
        strategy = metric or self.distance_strategy

        # Convert string metrics to DistanceStrategy enum if needed
        if isinstance(strategy, str):
            metric_mapping = {
                'COSINE': DistanceStrategy.COSINE,
                'L2': DistanceStrategy.EUCLIDEAN_DISTANCE,
                'EUCLIDEAN': DistanceStrategy.EUCLIDEAN_DISTANCE,
                'IP': DistanceStrategy.MAX_INNER_PRODUCT,
                'DOT': DistanceStrategy.DOT_PRODUCT,
                'DOT_PRODUCT': DistanceStrategy.DOT_PRODUCT,
                'MAX_INNER_PRODUCT': DistanceStrategy.MAX_INNER_PRODUCT
            }
            strategy = metric_mapping.get(strategy.upper(), DistanceStrategy.COSINE)

        return strategy

    async def similarity_search(
        self,
        query: str,
        collection: str = None,
        k: Optional[int] = None,
        limit: int = None,
        metadata_filters: Optional[Dict[str, Any]] = None,
        score_threshold: Optional[float] = None,
        metric: str = None,
        embedding_column: str = None,
        content_column: str = None,
        metadata_column: str = None,
        id_column: str = None,
        **kwargs
    ) -> List[SearchResult]:
        """
        Perform similarity search with optional threshold filtering.

        Args:
            query: The search query text
            collection: Collection name (optional, uses default if not provided)
            k: Number of results to return (alias for limit)
            limit: Maximum number of results to return
            metadata_filters: Dictionary of metadata filters to apply
            score_threshold: Minimum score threshold (results below threshold filtered out)
            metric: Distance metric to use ('COSINE', 'L2', 'IP')
            embedding_column: Name of the embedding column (for compatibility)
            content_column: Name of the content column (for compatibility)
            metadata_column: Name of the metadata column (for compatibility)
            id_column: Name of the ID column (for compatibility)
            **kwargs: Additional arguments

        Returns:
            List of SearchResult objects with content, metadata, score, and id
        """
        if not self._connected:
            await self.connection()

        collection = collection or self.collection_name

        if k and not limit:
            limit = k
        if not limit:
            limit = 10

        # Ensure collection exists
        if collection not in self._collections:
            self.logger.warning(f"Collection '{collection}' not found")
            return []

        collection_data = self._collections[collection]

        if collection_data['index'] is None or collection_data['index'].ntotal == 0:
            self.logger.warning(f"Collection '{collection}' is empty")
            return []

        # Generate query embedding
        query_embedding = await self._embed_.embed_query(query)

        # Convert to numpy array
        if isinstance(query_embedding, list):
            query_embedding = np.array(query_embedding, dtype=np.float32)
        elif not isinstance(query_embedding, np.ndarray):
            query_embedding = np.array(query_embedding, dtype=np.float32)

        # Ensure 2D array
        if query_embedding.ndim == 1:
            query_embedding = query_embedding.reshape(1, -1)

        # Normalize for cosine similarity
        if self.distance_strategy == DistanceStrategy.COSINE:
            faiss.normalize_L2(query_embedding)

        # Search FAISS index
        # For metadata filtering, we need to search more results and filter
        search_k = limit * 3 if metadata_filters else limit
        search_k = min(search_k, collection_data['index'].ntotal)

        distances, indices = collection_data['index'].search(query_embedding, search_k)

        # Convert distances to scores
        # FAISS returns distances, but we want similarity scores (higher is better)
        distances = distances[0]  # Get first row
        indices = indices[0]  # Get first row

        # Convert to scores based on metric
        if self.distance_strategy == DistanceStrategy.COSINE:
            # For cosine with normalized vectors, distance is actually similarity
            scores = distances
        elif self.distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            # Convert distance to similarity
            scores = 1.0 / (1.0 + distances)
        else:
            # For inner product, distance is already similarity-like
            scores = distances

        # Build results
        results = []
        for idx, score in zip(indices, scores):
            if idx == -1:  # FAISS returns -1 for empty slots
                continue

            # Get document ID
            doc_id = collection_data['idx_to_id'].get(idx)
            if doc_id is None:
                continue

            # Get document data
            content = collection_data['documents'].get(doc_id, "")
            metadata = collection_data['metadata'].get(doc_id, {})

            # Apply metadata filters
            if metadata_filters:
                match = True
                for key, value in metadata_filters.items():
                    if metadata.get(key) != value:
                        match = False
                        break
                if not match:
                    continue

            # Apply score threshold
            if score_threshold is not None:
                if self.distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
                    # For L2 distance, lower is better, so invert the check
                    if distances[np.where(indices == idx)[0][0]] > score_threshold:
                        continue
                else:
                    # For cosine and inner product, higher is better
                    if score < score_threshold:
                        continue

            # Create search result
            result = SearchResult(
                id=doc_id,
                content=content,
                metadata=metadata,
                score=float(score)
            )
            results.append(result)

            # Stop if we have enough results
            if len(results) >= limit:
                break

        self.logger.debug(
            f"Similarity search in collection '{collection}': "
            f"found {len(results)} results (limit={limit})"
        )

        return results

    async def asearch(
        self,
        query: str,
        collection: Optional[str] = None,
        k: Optional[int] = None,
        limit: Optional[int] = None,
        metadata_filters: Optional[Dict[str, Any]] = None,
        score_threshold: Optional[float] = None,
        metric: Optional[str] = None,
        embedding_column: Optional[str] = None,
        content_column: Optional[str] = None,
        metadata_column: Optional[str] = None,
        id_column: Optional[str] = None,
        **kwargs,
    ) -> List[SearchResult]:
        """Async alias for :meth:`similarity_search` to match store interface expectations."""

        return await self.similarity_search(
            query=query,
            collection=collection,
            k=k,
            limit=limit,
            metadata_filters=metadata_filters,
            score_threshold=score_threshold,
            metric=metric,
            embedding_column=embedding_column,
            content_column=content_column,
            metadata_column=metadata_column,
            id_column=id_column,
            **kwargs,
        )

    async def mmr_search(
        self,
        query: str,
        collection: str = None,
        k: int = 4,
        fetch_k: Optional[int] = None,
        lambda_mult: float = 0.5,
        metadata_filters: Optional[Dict[str, Any]] = None,
        score_threshold: Optional[float] = None,
        metric: str = None,
        embedding_column: str = None,
        content_column: str = None,
        metadata_column: str = None,
        id_column: str = None,
        **kwargs
    ) -> List[SearchResult]:
        """
        Perform MMR (Maximal Marginal Relevance) search for diversity.

        MMR balances relevance with diversity by selecting documents that are:
        1. Relevant to the query
        2. Diverse from already selected documents

        Args:
            query: The search query text
            collection: Collection name (optional, uses default if not provided)
            k: Number of results to return
            fetch_k: Number of candidates to fetch (default: k*3)
            lambda_mult: Diversity parameter (0-1)
                - 1.0 = maximum relevance (no diversity)
                - 0.0 = maximum diversity (no relevance)
            metadata_filters: Dictionary of metadata filters to apply
            score_threshold: Minimum score threshold
            metric: Distance metric to use ('COSINE', 'L2', 'IP')
            embedding_column: Name of the embedding column (for compatibility)
            content_column: Name of the content column (for compatibility)
            metadata_column: Name of the metadata column (for compatibility)
            id_column: Name of the ID column (for compatibility)
            **kwargs: Additional arguments

        Returns:
            List of SearchResult objects ordered by MMR score
        """
        if not self._connected:
            await self.connection()

        collection = collection or self.collection_name

        # Default to fetching 3x more candidates than final results
        if fetch_k is None:
            fetch_k = max(k * 3, 20)

        # Step 1: Get initial candidates using similarity search
        candidates = await self.similarity_search(
            query=query,
            collection=collection,
            limit=fetch_k,
            metadata_filters=metadata_filters,
            score_threshold=score_threshold,
            metric=metric,
            embedding_column=embedding_column,
            content_column=content_column,
            metadata_column=metadata_column,
            id_column=id_column,
            **kwargs
        )

        if len(candidates) <= k:
            # If we have fewer candidates than requested results, return all
            return candidates

        # Step 2: Get embeddings for MMR computation
        collection_data = self._collections[collection]
        candidate_embeddings = {}
        for result in candidates:
            embedding = collection_data['embeddings'].get(result.id)
            if embedding is not None:
                candidate_embeddings[result.id] = embedding

        # Step 3: Get query embedding
        query_embedding = await self._embed_.embed_query(query)
        if isinstance(query_embedding, list):
            query_embedding = np.array(query_embedding, dtype=np.float32)
        elif not isinstance(query_embedding, np.ndarray):
            query_embedding = np.array(query_embedding, dtype=np.float32)

        # Step 4: Run MMR algorithm
        selected_results = self._mmr_algorithm(
            query_embedding=query_embedding,
            candidates=candidates,
            candidate_embeddings=candidate_embeddings,
            k=k,
            lambda_mult=lambda_mult,
            metric=metric or self.distance_strategy
        )

        self.logger.info(
            f"MMR search in collection '{collection}': "
            f"selected {len(selected_results)} results from {len(candidates)} candidates "
            f"(λ={lambda_mult})"
        )

        return selected_results

    def _mmr_algorithm(
        self,
        query_embedding: np.ndarray,
        candidates: List[SearchResult],
        candidate_embeddings: Dict[str, np.ndarray],
        k: int,
        lambda_mult: float,
        metric: Union[str, DistanceStrategy]
    ) -> List[SearchResult]:
        """
        Core MMR algorithm implementation (same as PgVectorStore).

        Args:
            query_embedding: Query embedding vector
            candidates: List of candidate SearchResult objects
            candidate_embeddings: Dictionary mapping doc ID to embedding vector
            k: Number of results to select
            lambda_mult: MMR diversity parameter (0-1)
            metric: Distance metric to use

        Returns:
            List of selected SearchResult objects ordered by MMR score
        """
        if len(candidates) <= k:
            return candidates

        # Convert query embedding to numpy array
        if not isinstance(query_embedding, np.ndarray):
            query_embedding = np.array(query_embedding, dtype=np.float32)

        # Prepare data structures
        selected_indices = []
        remaining_indices = list(range(len(candidates)))

        # Step 1: Select the most relevant document first
        query_similarities = []
        for candidate in candidates:
            doc_embedding = candidate_embeddings.get(candidate.id)
            if doc_embedding is not None:
                similarity = self._compute_similarity(query_embedding, doc_embedding, metric)
                query_similarities.append(similarity)
            else:
                # Fallback to distance score if embedding not available
                query_similarities.append(1.0 / (1.0 + candidate.score))

        # Select the most similar document first
        best_idx = np.argmax(query_similarities)
        selected_indices.append(best_idx)
        remaining_indices.remove(best_idx)

        # Step 2: Iteratively select remaining documents using MMR
        for _ in range(min(k - 1, len(remaining_indices))):
            mmr_scores = []

            for idx in remaining_indices:
                candidate = candidates[idx]
                doc_embedding = candidate_embeddings.get(candidate.id)

                if doc_embedding is None:
                    # Fallback scoring if embedding not available
                    mmr_score = lambda_mult * query_similarities[idx]
                    mmr_scores.append(mmr_score)
                    continue

                # Relevance: similarity to query
                relevance = query_similarities[idx]

                # Diversity: maximum similarity to already selected documents
                max_similarity_to_selected = 0.0
                for selected_idx in selected_indices:
                    selected_candidate = candidates[selected_idx]
                    selected_embedding = candidate_embeddings.get(selected_candidate.id)

                    if selected_embedding is not None:
                        similarity = self._compute_similarity(doc_embedding, selected_embedding, metric)
                        max_similarity_to_selected = max(max_similarity_to_selected, similarity)

                # MMR formula: λ * relevance - (1-λ) * max_similarity_to_selected
                mmr_score = (
                    lambda_mult * relevance -
                    (1.0 - lambda_mult) * max_similarity_to_selected
                )
                mmr_scores.append(mmr_score)

            # Select document with highest MMR score
            if mmr_scores:
                best_remaining_idx = np.argmax(mmr_scores)
                best_idx = remaining_indices[best_remaining_idx]
                selected_indices.append(best_idx)
                remaining_indices.remove(best_idx)

        # Step 3: Return selected results with MMR scores in metadata
        selected_results = []
        for i, idx in enumerate(selected_indices):
            result = candidates[idx]
            # Add MMR ranking to metadata
            enhanced_metadata = dict(result.metadata)
            enhanced_metadata['mmr_rank'] = i + 1
            enhanced_metadata['mmr_lambda'] = lambda_mult
            enhanced_metadata['original_rank'] = idx + 1

            enhanced_result = SearchResult(
                id=result.id,
                content=result.content,
                metadata=enhanced_metadata,
                score=result.score
            )
            selected_results.append(enhanced_result)

        return selected_results

    def _compute_similarity(
        self,
        embedding1: np.ndarray,
        embedding2: np.ndarray,
        metric: Union[str, DistanceStrategy]
    ) -> float:
        """
        Compute similarity between two embeddings based on the specified metric.

        Args:
            embedding1: First embedding vector (numpy array or list)
            embedding2: Second embedding vector (numpy array or list)
            metric: Distance metric ('COSINE', 'L2', 'IP', etc.)

        Returns:
            Similarity score (higher = more similar)
        """
        # Convert to numpy arrays if needed
        if isinstance(embedding1, list):
            embedding1 = np.array(embedding1, dtype=np.float32)
        if isinstance(embedding2, list):
            embedding2 = np.array(embedding2, dtype=np.float32)

        # Ensure embeddings are numpy arrays
        if not isinstance(embedding1, np.ndarray):
            embedding1 = np.array(embedding1, dtype=np.float32)
        if not isinstance(embedding2, np.ndarray):
            embedding2 = np.array(embedding2, dtype=np.float32)

        # Ensure embeddings are 2D arrays for sklearn
        emb1 = embedding1.reshape(1, -1)
        emb2 = embedding2.reshape(1, -1)

        # Convert string metrics to DistanceStrategy enum if needed
        if isinstance(metric, str):
            metric_mapping = {
                'COSINE': DistanceStrategy.COSINE,
                'L2': DistanceStrategy.EUCLIDEAN_DISTANCE,
                'EUCLIDEAN': DistanceStrategy.EUCLIDEAN_DISTANCE,
                'IP': DistanceStrategy.MAX_INNER_PRODUCT,
                'DOT': DistanceStrategy.DOT_PRODUCT,
                'DOT_PRODUCT': DistanceStrategy.DOT_PRODUCT,
                'MAX_INNER_PRODUCT': DistanceStrategy.MAX_INNER_PRODUCT
            }
            strategy = metric_mapping.get(metric.upper(), DistanceStrategy.COSINE)
        else:
            strategy = metric

        if strategy == DistanceStrategy.COSINE:
            # Cosine similarity (returns similarity, not distance)
            similarity = cosine_similarity(emb1, emb2)[0, 0]

        elif strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            # Convert Euclidean distance to similarity
            distance = euclidean_distances(emb1, emb2)[0, 0]
            similarity = 1.0 / (1.0 + distance)

        elif strategy in [DistanceStrategy.MAX_INNER_PRODUCT, DistanceStrategy.DOT_PRODUCT]:
            # Dot product (inner product)
            similarity = np.dot(embedding1.flatten(), embedding2.flatten())

        else:
            # Default to cosine similarity
            similarity = cosine_similarity(emb1, emb2)[0, 0]
        return float(similarity)

    # Additional methods for compatibility

    def get_vector(self, metric_type: str = None, **kwargs):
        """
        Get the FAISS vector store (for compatibility).

        Args:
            metric_type: Distance metric type
            **kwargs: Additional arguments

        Returns:
            The FAISSStore instance itself
        """
        return self

    async def from_documents(
        self,
        documents: List[Document],
        collection: Union[str, None] = None,
        **kwargs
    ) -> 'FAISSStore':
        """
        Create Vector Store from Documents.

        Args:
            documents: List of Documents
            collection: Collection Name
            **kwargs: Additional Arguments

        Returns:
            The FAISSStore instance
        """
        await self.add_documents(documents, collection=collection, **kwargs)
        return self

    # Persistence methods

    def save(self, file_path: Union[str, Path]) -> None:
        """
        Save the FAISS store to disk.

        Args:
            file_path: Path to save the store
        """
        file_path = Path(file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        # Prepare data for saving
        save_data = {
            'collections': {},
            'config': {
                'collection_name': self.collection_name,
                'distance_strategy': self.distance_strategy.value,
                'index_type': self.index_type,
                'dimension': self.dimension,
            }
        }

        # Save each collection
        for coll_name, coll_data in self._collections.items():
            # Save FAISS index
            index_path = file_path.parent / f"{file_path.stem}_{coll_name}.index"
            if coll_data['index'] is not None:
                faiss.write_index(coll_data['index'], str(index_path))

            # Save metadata and documents
            save_data['collections'][coll_name] = {
                'documents': coll_data['documents'],
                'metadata': coll_data['metadata'],
                'embeddings': {k: v.tolist() for k, v in coll_data['embeddings'].items()},
                'id_to_idx': coll_data['id_to_idx'],
                'idx_to_id': coll_data['idx_to_id'],
                'dimension': coll_data['dimension'],
                'is_trained': coll_data['is_trained'],
                'index_path': str(index_path),
            }

        # Save to pickle
        with open(file_path, 'wb') as f:
            pickle.dump(save_data, f)

        self.logger.info(f"Saved FAISS store to {file_path}")

    def load(self, file_path: Union[str, Path]) -> None:
        """
        Load the FAISS store from disk.

        Args:
            file_path: Path to load the store from
        """
        file_path = Path(file_path)

        with open(file_path, 'rb') as f:
            save_data = pickle.load(f)

        # Restore config
        config = save_data.get('config', {})
        self.collection_name = config.get('collection_name', self.collection_name)

        # Restore collections
        for coll_name, coll_data in save_data['collections'].items():
            self._initialize_collection(coll_name)

            # Load FAISS index
            index_path = coll_data['index_path']
            if Path(index_path).exists():
                index = faiss.read_index(index_path)

                self._collections[coll_name]['index'] = index

            # Load metadata and documents
            self._collections[coll_name]['documents'] = coll_data['documents']
            self._collections[coll_name]['metadata'] = coll_data['metadata']
            self._collections[coll_name]['embeddings'] = {
                k: np.array(v, dtype=np.float32)
                for k, v in coll_data['embeddings'].items()
            }
            self._collections[coll_name]['id_to_idx'] = coll_data['id_to_idx']
            self._collections[coll_name]['idx_to_id'] = coll_data['idx_to_id']
            self._collections[coll_name]['dimension'] = coll_data['dimension']
            self._collections[coll_name]['is_trained'] = coll_data['is_trained']

        self.logger.info(f"Loaded FAISS store from {file_path}")

    def __str__(self) -> str:
        return f"FAISSStore(collection={self.collection_name}, index_type={self.index_type})"

    def __repr__(self) -> str:
        return (
            f"<FAISSStore(collection='{self.collection_name}', "
            f"index_type='{self.index_type}', "
            f"distance_strategy='{self.distance_strategy.value}', "
            "cpu_only=True)>"
        )

    async def delete_documents(
        self,
        document_ids: List[str],
        collection: str = None,
        **kwargs
    ) -> None:
        """
        Delete documents by their IDs from the FAISS store.

        Args:
            document_ids: List of document IDs to delete
            collection: Collection name (optional, uses default if not provided)
            **kwargs: Additional arguments
        """
        if not self._connected:
            await self.connection()

        collection = collection or self.collection_name

        # Ensure collection exists
        if collection not in self._collections:
            self.logger.warning(f"Collection '{collection}' not found")
            return

        collection_data = self._collections[collection]

        for doc_id in document_ids:
            idx = collection_data['id_to_idx'].get(doc_id)
            if idx is not None:
                # Remove from FAISS index
                # Note: FAISS does not support direct deletion; we mark as deleted
                # Here we simply ignore the vector in searches by removing mappings
                del collection_data['documents'][doc_id]
                del collection_data['metadata'][doc_id]
                del collection_data['embeddings'][doc_id]
                del collection_data['id_to_idx'][doc_id]
                del collection_data['idx_to_id'][idx]

        self.logger.info(
            f"✅ Successfully deleted {len(document_ids)} documents from collection '{collection}'"
        )

    async def delete_documents_by_filter(self, filter_func, collection: str = None, **kwargs) -> None:
        """
        Delete documents that match a filter function from the FAISS store.

        Args:
            filter_func: A function that takes metadata and returns True if the document should be deleted
            collection: Collection name (optional, uses default if not provided)
            **kwargs: Additional arguments
        """
        if not self._connected:
            await self.connection()

        collection = collection or self.collection_name

        # Ensure collection exists
        if collection not in self._collections:
            self.logger.warning(f"Collection '{collection}' not found")
            return

        collection_data = self._collections[collection]

        to_delete_ids = []
        for doc_id, metadata in collection_data['metadata'].items():
            if filter_func(metadata):
                to_delete_ids.append(doc_id)

        await self.delete_documents(to_delete_ids, collection=collection, **kwargs)



"""
Conversational Form Dialog - One prompt per field, text-based interaction.

Uses BotBuilder's native prompts instead of Adaptive Cards.
Useful for:
- More natural conversation flow
- Channels with limited Adaptive Card support
- Complex fields requiring contextual help
- Accessibility considerations
"""
from typing import Any, Callable, Awaitable, Dict, List, Optional
from botbuilder.dialogs import (
    WaterfallDialog,
    WaterfallStepContext,
    DialogTurnResult,
    DialogTurnStatus,
)
from botbuilder.dialogs.prompts import (
    TextPrompt,
    NumberPrompt,
    ConfirmPrompt,
    ChoicePrompt,
    DateTimePrompt,
    PromptOptions,
    PromptValidatorContext,
)
from botbuilder.dialogs.choices import Choice, FoundChoice
from botbuilder.core import TurnContext, MessageFactory
from .base import BaseFormDialog
from ....dialogs.models import (
    FormDefinition,
    FormField,
    FieldType,
    ValidationRule,
)
from ..validator import FormValidator


class ConversationalFormDialog(BaseFormDialog):
    """
    Conversational form using native BotBuilder prompts.

    Each field becomes a separate prompt in the waterfall.
    Supports:
    - TextPrompt for text fields
    - NumberPrompt for numeric fields
    - ChoicePrompt for single choice
    - ConfirmPrompt for boolean/toggle
    - DateTimePrompt for dates

    Features:
    - Field-level validation with retry
    - Contextual help messages
    - Skip optional fields
    - Back navigation (restart)

    Flow:
    1. "What is your first name?"
    2. User types: "John"
    3. "What is your email?"
    4. User types: "john@example.com"
    5. ... continues for each field ...
    6. "All done! Here's your summary..."
    """

    def __init__(
        self,
        form: FormDefinition,
        dialog_id: str = None,
        show_progress: bool = True,
        allow_skip_optional: bool = True,
        **kwargs,  # Accept but ignore extra kwargs for backwards compatibility
    ):
        super().__init__(form=form, dialog_id=dialog_id)

        self.show_progress = show_progress
        self.allow_skip_optional = allow_skip_optional

        # Flatten all fields from all sections
        self._fields: List[FormField] = []
        for section in form.sections:
            self._fields.extend(section.fields)

        # Register prompts with validators
        self._register_prompts()

        # Build waterfall steps
        steps = self._build_steps()

        self.add_dialog(
            WaterfallDialog(
                f"{self.form.form_id}_waterfall",
                steps,
            )
        )

        self.initial_dialog_id = f"{self.form.form_id}_waterfall"

    def _register_prompts(self):
        """Register prompt dialogs for each field type."""
        # Text prompt with validation
        self.add_dialog(
            TextPrompt(
                "TextPrompt",
                self._text_validator,
            )
        )

        # Number prompt
        self.add_dialog(
            NumberPrompt(
                "NumberPrompt",
                self._number_validator,
            )
        )

        # Choice prompt
        self.add_dialog(ChoicePrompt("ChoicePrompt"))

        # Confirm prompt (for toggles/booleans)
        self.add_dialog(ConfirmPrompt("ConfirmPrompt"))

        # DateTime prompt
        self.add_dialog(DateTimePrompt("DateTimePrompt"))

        # Email-specific text prompt
        self.add_dialog(
            TextPrompt(
                "EmailPrompt",
                self._email_validator,
            )
        )

        # URL-specific text prompt
        self.add_dialog(
            TextPrompt(
                "UrlPrompt",
                self._url_validator,
            )
        )

    def _build_steps(self) -> List[Callable]:
        """Build waterfall steps: intro + one per field + summary."""
        steps = []

        # Introduction step
        steps.append(self.intro_step)

        # One step per field
        for i, field in enumerate(self._fields):
            steps.append(self._create_field_prompt_step(i))
            steps.append(self._create_field_process_step(i))

        # Summary/completion step
        steps.append(self.summary_step)

        return steps

    # =========================================================================
    # Waterfall Steps
    # =========================================================================

    async def intro_step(
        self,
        step_context: WaterfallStepContext,
    ) -> DialogTurnResult:
        """Introduction step with form title and instructions."""

        intro_message = (
            f"📋 **{self.form.title}**\n\n"
            f"I'll ask you a few questions to collect the information needed. "
        )

        if self.allow_skip_optional:
            intro_message += "Type 'skip' to skip optional fields. "

        intro_message += "Type 'cancel' at any time to stop.\n"

        await step_context.context.send_activity(
            MessageFactory.text(intro_message)
        )

        return await step_context.next(None)

    def _create_field_prompt_step(self, field_index: int) -> Callable:
        """Create step that prompts for a field."""

        async def prompt_step(
            step_context: WaterfallStepContext,
        ) -> DialogTurnResult:
            field = self._fields[field_index]

            # Build prompt message
            prompt_text = self._build_prompt_message(field, field_index)

            # Get appropriate prompt type and options
            prompt_type, options = self._get_prompt_config(field, prompt_text)

            return await step_context.prompt(prompt_type, options)

        prompt_step.__name__ = f"prompt_{field_index}"
        return prompt_step

    def _create_field_process_step(self, field_index: int) -> Callable:
        """Create step that processes field response."""

        async def process_step(
            step_context: WaterfallStepContext,
        ) -> DialogTurnResult:
            field = self._fields[field_index]
            result = step_context.result

            # Check for cancel
            if isinstance(result, str) and result.lower() == 'cancel':
                return await self.handle_cancel(step_context)

            # Check for skip (optional fields only)
            if isinstance(result, str) and result.lower() == 'skip':
                if not field.required and self.allow_skip_optional:
                    await step_context.context.send_activity(
                        MessageFactory.text(f"Skipped {field.label or field.name}.")
                    )
                    return await step_context.next(None)
                else:
                    await step_context.context.send_activity(
                        MessageFactory.text(
                            f"⚠️ {field.label or field.name} is required and cannot be skipped."
                        )
                    )
                    # Re-prompt
                    return await step_context.replace_dialog(self.id)

            # Process and store the value
            processed_value = self._process_result(result, field)

            form_data = self.get_form_data(step_context)
            form_data[field.name] = processed_value
            self.set_form_data(step_context, form_data)

            # Acknowledge (optional, can be verbose)
            if field.field_type == FieldType.TOGGLE:
                ack = "Yes" if processed_value else "No"
            else:
                ack = str(processed_value)

            # Only show for certain types to avoid clutter
            if field.field_type in (FieldType.CHOICE, FieldType.TOGGLE):
                await step_context.context.send_activity(
                    MessageFactory.text(f"✓ {field.label}: {ack}")
                )

            return await step_context.next(processed_value)

        process_step.__name__ = f"process_{field_index}"
        return process_step

    async def summary_step(
        self,
        step_context: WaterfallStepContext,
    ) -> DialogTurnResult:
        """Show summary and complete."""

        form_data = self.get_form_data(step_context)

        # Build summary message
        summary = self._build_summary_message(form_data)

        await step_context.context.send_activity(
            MessageFactory.text(summary)
        )

        # Complete the form
        return await self.handle_complete(step_context, form_data)

    # =========================================================================
    # Prompt Configuration
    # =========================================================================

    def _build_prompt_message(
        self,
        field: FormField,
        field_index: int,
    ) -> str:
        """Build the prompt message for a field."""
        parts = []

        # Progress indicator
        if self.show_progress:
            total = len(self._fields)
            parts.append(f"({field_index + 1}/{total})")

        # Field label/question
        label = field.label or field.name.replace("_", " ").title()

        if field.required:
            parts.append(f"**{label}** (required)")
        else:
            parts.append(f"**{label}** (optional)")

        prompt = " ".join(parts)

        # Add description as hint
        if field.description:
            prompt += f"\n_{field.description}_"

        # Add placeholder as example
        if field.placeholder:
            prompt += f"\n_Example: {field.placeholder}_"

        return prompt

    def _get_prompt_config(
        self,
        field: FormField,
        prompt_text: str,
    ) -> tuple:
        """Get prompt type and options for a field."""

        field_type = field.field_type

        # Text-based fields
        if field_type == FieldType.TEXT:
            return "TextPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
                retry_prompt=MessageFactory.text(
                    f"Please enter a valid value for {field.label or field.name}."
                ),
            )

        elif field_type == FieldType.TEXTAREA:
            return "TextPrompt", PromptOptions(
                prompt=MessageFactory.text(
                    f"{prompt_text}\n(You can write multiple lines)"
                ),
            )

        elif field_type == FieldType.EMAIL:
            return "EmailPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
                retry_prompt=MessageFactory.text(
                    "Please enter a valid email address (e.g., name@example.com)."
                ),
            )

        elif field_type == FieldType.URL:
            return "UrlPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
                retry_prompt=MessageFactory.text(
                    "Please enter a valid URL (e.g., https://example.com)."
                ),
            )

        # Number field
        elif field_type == FieldType.NUMBER:
            return "NumberPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
                retry_prompt=MessageFactory.text(
                    "Please enter a valid number."
                ),
            )

        # Date fields
        elif field_type in (FieldType.DATE, FieldType.DATETIME):
            return "DateTimePrompt", PromptOptions(
                prompt=MessageFactory.text(
                    f"{prompt_text}\n(e.g., tomorrow, next Monday, 2025-01-15)"
                ),
                retry_prompt=MessageFactory.text(
                    "I couldn't understand that date. Try something like 'next Friday' or '2025-01-15'."
                ),
            )

        # Single choice
        elif field_type == FieldType.CHOICE:
            choices = self._build_choices_list(field.choices)
            return "ChoicePrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
                choices=choices,
                retry_prompt=MessageFactory.text(
                    f"Please choose one of the options: {', '.join(c.value for c in choices)}"
                ),
            )

        # Multi-choice (use text prompt, parse manually)
        elif field_type == FieldType.MULTICHOICE:
            options_text = ", ".join(self._extract_choice_values(field.choices))
            return "TextPrompt", PromptOptions(
                prompt=MessageFactory.text(
                    f"{prompt_text}\nOptions: {options_text}\n(Enter choices separated by commas)"
                ),
            )

        # Toggle/Boolean
        elif field_type == FieldType.TOGGLE:
            return "ConfirmPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
            )

        # Default to text
        else:
            return "TextPrompt", PromptOptions(
                prompt=MessageFactory.text(prompt_text),
            )

    def _build_choices_list(
        self,
        choices: Optional[List],
    ) -> List[Choice]:
        """Build Choice objects for ChoicePrompt."""
        if not choices:
            return []

        result = []
        for choice in choices:
            if isinstance(choice, str):
                result.append(Choice(value=choice))
            elif isinstance(choice, dict):
                result.append(Choice(
                    value=choice.get("value", choice.get("title", "")),
                    action=None,
                    synonyms=[choice.get("title")] if choice.get("title") != choice.get("value") else None,
                ))

        return result

    def _extract_choice_values(
        self,
        choices: Optional[List],
    ) -> List[str]:
        """Extract string values from choices."""
        if not choices:
            return []

        values = []
        for choice in choices:
            if isinstance(choice, str):
                values.append(choice)
            elif isinstance(choice, dict):
                values.append(choice.get("value", choice.get("title", "")))

        return values

    def _process_result(
        self,
        result: Any,
        field: FormField,
    ) -> Any:
        """Process prompt result into storage format."""

        # ChoicePrompt returns FoundChoice
        if isinstance(result, FoundChoice):
            return result.value

        # ConfirmPrompt returns bool
        if isinstance(result, bool):
            return result

        # DateTimePrompt returns list of resolutions
        if isinstance(result, list) and result:
            if hasattr(result[0], 'value'):
                return result[0].value
            return result[0]

        # Multi-choice: parse comma-separated
        if field.field_type == FieldType.MULTICHOICE and isinstance(result, str):
            return [v.strip() for v in result.split(',') if v.strip()]

        return result

    # =========================================================================
    # Validators
    # =========================================================================

    async def _text_validator(
        self,
        prompt_context: PromptValidatorContext,
    ) -> bool:
        """Validate text input."""
        value = prompt_context.recognized.value

        if not value or not value.strip():
            return False

        # Allow cancel/skip commands
        if value.lower() in ('cancel', 'skip'):
            return True

        return True

    async def _email_validator(
        self,
        prompt_context: PromptValidatorContext,
    ) -> bool:
        """Validate email input."""
        import re

        value = prompt_context.recognized.value

        if not value:
            return False

        # Allow cancel/skip
        if value.lower() in ('cancel', 'skip'):
            return True

        pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
        return bool(re.match(pattern, value.strip()))

    async def _url_validator(
        self,
        prompt_context: PromptValidatorContext,
    ) -> bool:
        """Validate URL input."""
        import re

        value = prompt_context.recognized.value

        if not value:
            return False

        # Allow cancel/skip
        if value.lower() in ('cancel', 'skip'):
            return True

        pattern = r'^https?://[^\s/$.?#].[^\s]*$'
        return bool(re.match(pattern, value.strip()))

    async def _number_validator(
        self,
        prompt_context: PromptValidatorContext,
    ) -> bool:
        """Validate number input."""
        # NumberPrompt already validates, just ensure it succeeded
        return prompt_context.recognized.succeeded

    # =========================================================================
    # Summary
    # =========================================================================

    def _build_summary_message(
        self,
        form_data: Dict[str, Any],
    ) -> str:
        """Build a text summary of collected data."""
        lines = [
            f"✅ **{self.form.title} - Complete**",
            "",
            "Here's what you submitted:",
            "",
        ]

        for section in self.form.sections:
            lines.append(f"**{section.title}**")

            for field in section.fields:
                value = form_data.get(field.name)

                if value is None:
                    display = "_skipped_"
                elif isinstance(value, bool):
                    display = "Yes" if value else "No"
                elif isinstance(value, list):
                    display = ", ".join(str(v) for v in value)
                else:
                    display = str(value)

                lines.append(f"• {field.label or field.name}: {display}")

            lines.append("")

        return "\n".join(lines)




import pytest
import os
from parrot.tools.fred_api import FredAPITool

@pytest.mark.asyncio
async def test_fred_api_tool_fedfunds():
    """Test fetching FEDFUNDS series from FredAPITool."""
    tool = FredAPITool()
    
    # Check if API key is present in env, otherwise skip or warn
    if not os.environ.get("FRED_API_KEY"):
        pytest.skip("FRED_API_KEY not found in environment")

    result = await tool._execute(series_id="FEDFUNDS", params={"limit": 1})
    
    assert result.success is True
    assert result.status == "success"
    assert result.result is not None
    assert "observations" in result.result

@pytest.mark.asyncio
async def test_fred_api_tool_no_key():
    """Test FredAPITool error handling when no API key is provided."""
    # Temporarily unset env var if it exists
    original_key = os.environ.get("FRED_API_KEY")
    if original_key:
        del os.environ["FRED_API_KEY"]
    
    try:
        tool = FredAPITool()
        result = await tool._execute(series_id="FEDFUNDS")
        assert result.success is False
        assert "API Key not found" in result.error
    finally:
        if original_key:
            os.environ["FRED_API_KEY"] = original_key




"""ClickHouse database source for DatabaseToolkit.

Implements ``AbstractDatabaseSource`` for ClickHouse OLAP database using the
asyncdb ``clickhouse`` driver. Queries ``system.columns`` for metadata discovery.
Inherits SQL validation from the base class via ``sqlglot_dialect = "clickhouse"``.

Part of FEAT-062 — DatabaseToolkit.
"""
from __future__ import annotations

import logging
import time
from typing import Any

from parrot.tools.database.base import (
    AbstractDatabaseSource,
    ColumnMeta,
    MetadataResult,
    QueryResult,
    RowResult,
    TableMeta,
    _validate_sql_identifier,
)
from parrot.tools.database.sources import register_source


@register_source("clickhouse")
class ClickHouseSource(AbstractDatabaseSource):
    """ClickHouse OLAP database source.

    Uses the asyncdb ``clickhouse`` driver. Validates queries with the
    ``clickhouse`` sqlglot dialect. Discovers schema via system.columns.
    """

    driver = "clickhouse"
    sqlglot_dialect = "clickhouse"

    def __init__(self) -> None:
        """Initialize ClickHouseSource with a logger."""
        self.logger = logging.getLogger("Parrot.Toolkits.Database.ClickHouse")

    async def get_default_credentials(self) -> dict[str, Any]:
        """Return default ClickHouse credentials.

        Returns:
            Empty dict (no default ClickHouse credentials configured).
        """
        from parrot.interfaces.database import get_default_credentials
        dsn = get_default_credentials("clickhouse")
        return {"dsn": dsn} if dsn else {}

    async def get_metadata(
        self,
        credentials: dict[str, Any],
        tables: list[str] | None = None,
    ) -> MetadataResult:
        """Discover ClickHouse schema via system.columns.

        Args:
            credentials: Connection credentials (``database`` key for DB name).
            tables: Optional list of specific table names to inspect.

        Returns:
            MetadataResult with table and column metadata.
        """
        self.logger.debug("get_metadata called, tables=%s", tables)
        # Validate table names before acquiring a connection (fail-fast on injection)
        if tables:
            validated_tables = [_validate_sql_identifier(t, "table name") for t in tables]
        else:
            validated_tables = None

        database = credentials.get("database", credentials.get("db", "default"))
        dsn = credentials.get("dsn")
        params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("clickhouse", dsn, params)
        async with await db.connection() as conn:
            if validated_tables:
                placeholders = ", ".join([f"'{t}'" for t in validated_tables])
                sql = f"""
                    SELECT database, table, name AS column_name, type AS data_type,
                           default_expression AS column_default,
                           is_in_primary_key
                    FROM system.columns
                    WHERE database = %(db)s AND table IN ({placeholders})
                    ORDER BY table, position
                """
            else:
                sql = """
                    SELECT database, table, name AS column_name, type AS data_type,
                           default_expression AS column_default,
                           is_in_primary_key
                    FROM system.columns
                    WHERE database = %(db)s
                    ORDER BY table, position
                """
            rows = await conn.fetch_all(sql, db=database)

        tables_map: dict[str, TableMeta] = {}
        for row in (rows or []):
            row_dict = dict(row) if not isinstance(row, dict) else row
            table_name = row_dict.get("table", "")
            if table_name not in tables_map:
                tables_map[table_name] = TableMeta(
                    name=table_name,
                    schema_name=row_dict.get("database", database),
                )
            col = ColumnMeta(
                name=row_dict.get("column_name", ""),
                data_type=row_dict.get("data_type", "unknown"),
                nullable="Nullable" in row_dict.get("data_type", ""),
                primary_key=bool(row_dict.get("is_in_primary_key", 0)),
                default=row_dict.get("column_default") or None,
            )
            tables_map[table_name].columns.append(col)

        return MetadataResult(driver=self.driver, tables=list(tables_map.values()))

    async def query(
        self,
        credentials: dict[str, Any],
        sql: str,
        params: dict[str, Any] | None = None,
    ) -> QueryResult:
        """Execute a ClickHouse SQL query and return all results.

        Args:
            credentials: Connection credentials.
            sql: ClickHouse SQL query string.
            params: Optional query parameters.

        Returns:
            QueryResult with rows and execution metadata.
        """
        self.logger.debug("query called: %s", sql[:100])
        start = time.monotonic()
        dsn = credentials.get("dsn")
        conn_params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("clickhouse", dsn, conn_params)
        async with await db.connection() as conn:
            if params:
                rows = await conn.fetch_all(sql, **params)
            else:
                rows = await conn.fetch_all(sql)

        elapsed_ms = (time.monotonic() - start) * 1000
        rows_list = [dict(r) if not isinstance(r, dict) else r for r in (rows or [])]
        columns = list(rows_list[0].keys()) if rows_list else []

        return QueryResult(
            driver=self.driver,
            rows=rows_list,
            row_count=len(rows_list),
            columns=columns,
            execution_time_ms=round(elapsed_ms, 3),
        )

    async def query_row(
        self,
        credentials: dict[str, Any],
        sql: str,
        params: dict[str, Any] | None = None,
    ) -> RowResult:
        """Execute a ClickHouse SQL query and return the first row.

        Args:
            credentials: Connection credentials.
            sql: ClickHouse SQL query string.
            params: Optional query parameters.

        Returns:
            RowResult with a single row or found=False.
        """
        self.logger.debug("query_row called: %s", sql[:100])
        start = time.monotonic()
        dsn = credentials.get("dsn")
        conn_params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("clickhouse", dsn, conn_params)
        async with await db.connection() as conn:
            if params:
                row = await conn.fetch_one(sql, **params)
            else:
                row = await conn.fetch_one(sql)

        elapsed_ms = (time.monotonic() - start) * 1000
        row_dict = dict(row) if row and not isinstance(row, dict) else row

        return RowResult(
            driver=self.driver,
            row=row_dict,
            found=row_dict is not None,
            execution_time_ms=round(elapsed_ms, 3),
        )



import os
from typing import List, Optional, Union, Dict, Any
import contextlib
from pathlib import Path
from urllib.parse import quote, unquote
import asyncio
import aiofiles
from tqdm import tqdm
from io import BytesIO
import httpx
import aiohttp
import pandas as pd
# Microsoft Graph SDK
from msgraph.generated.models.drive_item import DriveItem
from msgraph.generated.models.folder import Folder
from msgraph.generated.models.file import File
from msgraph.generated.models.upload_session import UploadSession
from msgraph.generated.drives.item.items.item.create_upload_session.create_upload_session_post_request_body import (
    CreateUploadSessionPostRequestBody
)
from msgraph.generated.models.drive_item_uploadable_properties import DriveItemUploadableProperties
from .o365 import O365Client


class OneDriveClient(O365Client):
    """
    OneDrive Client - Migrated to Microsoft Graph SDK

    Uses Microsoft Graph SDK for all OneDrive operations.

    Interface for Managing connections to OneDrive resources.

    Methods:
        file_list: Lists files in a specified OneDrive folder.
        file_search: Searches for files matching a query.
        file_download: Downloads a single file by its item ID.
        download_files: Downloads multiple files provided as a list of dictionaries containing file info.
        folder_download: Downloads a folder and its contents recursively.
        file_delete: Deletes a file or folder by its item ID.
        upload_files: Uploads multiple files to a specified OneDrive folder.
        upload_file: Uploads a single file to OneDrive.
        upload_folder: Uploads a local folder and its contents to OneDrive recursively.
        download_excel_file: Downloads Excel files and optionally converts to pandas DataFrame.
        upload_dataframe_as_excel: Uploads pandas DataFrame as Excel file to OneDrive.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # OneDrive-specific properties
        self.directory: Optional[str] = None
        self.filename: Optional[str] = None
        self._srcfiles: List = []
        self._destination: List = []

        # Upload settings
        self.small_file_threshold = 4 * 1024 * 1024  # 4 MB
        self.chunk_size = 10 * 1024 * 1024  # 10 MB

        # Cached OneDrive objects
        self._drive_id: Optional[str] = None
        self._drive_info: Optional[DriveItem] = None

    def connection(self):
        """
        Establish OneDrive connection using the migrated O365Client.

        This replaces the old office365-rest-python-client authentication
        with Microsoft Graph SDK authentication.
        """
        # Use the parent O365Client connection method
        super().connection()

        self.logger.info("OneDrive connection established successfully")
        return self

    async def verify_onedrive_access(self):
        """Verify OneDrive access and cache drive info."""
        try:
            # Resolve and cache drive info
            self._drive_info = await self._resolve_drive()
            self.logger.info(f"OneDrive accessible: {self._drive_info.name or 'Personal OneDrive'}")

        except Exception as e:
            self.logger.error(f"OneDrive access verification failed: {e}")
            raise RuntimeError(f"OneDrive access verification failed: {e}") from e

    async def _resolve_drive(self) -> DriveItem:
        """Resolve OneDrive using Graph API."""
        if self._drive_info:
            return self._drive_info

        try:
            # Get user's personal OneDrive
            drive = await self.graph_client.me.drive.get()

            if drive and drive.id:
                self._drive_id = drive.id
                self._drive_info = drive
                self.logger.info(f"OneDrive resolved: {drive.name or drive.id}")
                return drive
            else:
                raise RuntimeError("Could not resolve OneDrive")

        except Exception as e:
            raise RuntimeError(f"Failed to resolve OneDrive: {e}") from e

    async def _ensure_folder(self, folder_path: str, create: bool = True) -> DriveItem:
        """Ensure folder exists in OneDrive using Graph API."""
        drive_info = await self._resolve_drive()
        drive_id = drive_info.id

        folder_path = (folder_path or "").strip("/")
        if not folder_path:
            # Return root folder
            root = await self.graph_client.drives.by_drive_id(drive_id).root.get()
            return root

        # Try to resolve existing folder
        try:
            folder_item = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(f"root:/{folder_path}:").get()
            if folder_item:
                return folder_item
        except Exception:
            if not create:
                raise

        # Create folder recursively
        root = await self.graph_client.drives.by_drive_id(drive_id).root.get()
        parent_id = root.id

        for segment in [s for s in folder_path.split("/") if s]:
            # Check if segment already exists
            children = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(parent_id).children.get()

            existing_folder = None
            if children and children.value:
                for child in children.value:
                    if child.name == segment and child.folder:
                        existing_folder = child
                        break

            if existing_folder:
                parent_id = existing_folder.id
                continue

            # Create new folder
            new_folder = DriveItem()
            new_folder.name = segment
            new_folder.folder = Folder()
            new_folder.additional_data = {
                "@microsoft.graph.conflictBehavior": "replace"
            }

            created = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(parent_id).children.post(new_folder)
            parent_id = created.id
            self.logger.info(f"Created folder: {segment}")

        # Return the final folder
        final_folder = await self.graph_client.drives.by_drive_id(drive_id)\
            .items.by_drive_item_id(parent_id).get()
        return final_folder

    async def file_list(self, folder_path: str = None) -> List[dict]:
        """
        List files in a given OneDrive folder using Microsoft Graph API.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            if folder_path:
                # Get specific folder
                folder_path = folder_path.strip("/")
                try:
                    folder_item = await self.graph_client.drives.by_drive_id(drive_id)\
                        .items.by_drive_item_id(f"root:/{folder_path}:").get()
                except Exception as e:
                    raise RuntimeError(f"Folder '{folder_path}' not found: {e}") from e
            else:
                # Get root folder
                folder_item = await self.graph_client.drives.by_drive_id(drive_id).root.get()

            # Get children
            children = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(folder_item.id).children.get()

            file_list = []
            if children and children.value:
                for item in children.value:
                    file_info = {
                        "name": item.name,
                        "id": item.id,
                        "webUrl": item.web_url,
                        "path": self._get_item_path_from_item(item),
                        "isFolder": item.folder is not None,
                        "size": item.size or 0,
                        "modified": item.last_modified_date_time.isoformat() if item.last_modified_date_time else None
                    }
                    file_list.append(file_info)

            return file_list

        except Exception as err:
            self.logger.error(f"Error listing files: {err}")
            raise RuntimeError(f"Error listing files: {err}") from err

    async def file_search(self, search_query: str) -> List[dict]:
        """
        Search for files in OneDrive matching the search query using Microsoft Graph API.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Use Graph API search
            search_results = await self.graph_client.drives.by_drive_id(drive_id)\
                .search_with_q(search_query).get()

            results = []
            if search_results and search_results.value:
                for item in search_results.value:
                    if item.file:  # Only include files, not folders
                        file_info = {
                            "name": item.name,
                            "id": item.id,
                            "webUrl": item.web_url,
                            "path": self._get_item_path_from_item(item),
                            "isFolder": False,
                            "size": item.size or 0,
                            "modified": item.last_modified_date_time.isoformat() if item.last_modified_date_time else None   # noqa
                        }
                        results.append(file_info)

            return results

        except Exception as err:
            self.logger.error(f"Error searching files: {err}")
            raise RuntimeError(f"Error searching files: {err}") from err

    async def file_download(self, item_id: str, destination: Path) -> str:
        """
        Download a file from OneDrive by item ID using Microsoft Graph API.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Get item info
            item = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(item_id).get()

            if not item.file:
                raise RuntimeError(f"Item {item_id} is not a file")

            self.logger.info(f"Downloading {item.name} to {destination}")

            # Get download URL if available
            download_url = ""
            try:
                add = getattr(item, "additional_data", {}) or {}
                download_url = add.get("@microsoft.graph.downloadUrl", "") or ""
            except Exception:
                download_url = ""

            # Ensure destination directory exists
            destination.parent.mkdir(parents=True, exist_ok=True)

            if download_url:
                # Stream via downloadUrl
                async with httpx.AsyncClient(follow_redirects=True, timeout=None) as client:
                    async with client.stream("GET", download_url) as resp:
                        resp.raise_for_status()
                        async with aiofiles.open(destination, "wb") as f:
                            async for chunk in resp.aiter_bytes(1 << 20):  # 1 MiB
                                await f.write(chunk)
            else:
                # Fallback: GET /content via Graph
                content = await self.graph_client.drives.by_drive_id(drive_id)\
                    .items.by_drive_item_id(item_id).content.get()
                async with aiofiles.open(destination, "wb") as f:
                    await f.write(content)

            self.logger.info(f"Downloaded {item.name} successfully")
            return str(destination)

        except Exception as err:
            self.logger.error(f"Error downloading file {item_id}: {err}")
            raise RuntimeError(f"Error downloading file {item_id}: {err}") from err

    async def download_files(self, items: List[dict], destination_folder: Path) -> List[str]:
        """
        Download multiple files from OneDrive using Microsoft Graph API.
        """
        downloaded_files = []
        destination_folder = Path(destination_folder)
        destination_folder.mkdir(parents=True, exist_ok=True)

        for item in items:
            try:
                item_id = item.get("id")
                file_name = item.get("name")
                if not item_id or not file_name:
                    self.logger.warning(f"Skipping invalid item: {item}")
                    continue

                destination = destination_folder / file_name
                downloaded_path = await self.file_download(item_id, destination)
                downloaded_files.append(downloaded_path)

            except Exception as e:
                self.logger.error(f"Failed to download {item.get('name', 'unknown')}: {e}")
                continue

        return downloaded_files

    async def folder_download(self, folder_id: str, destination_folder: Path) -> bool:
        """
        Download a folder and its contents from OneDrive using Microsoft Graph API.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Get folder info
            folder_item = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(folder_id).get()

            if not folder_item.folder:
                raise RuntimeError(f"Item {folder_id} is not a folder")

            await self._download_folder_recursive(drive_id, folder_item, destination_folder)
            return True

        except Exception as err:
            self.logger.error(f"Error downloading folder {folder_id}: {err}")
            raise RuntimeError(f"Error downloading folder {folder_id}: {err}") from err

    async def _download_folder_recursive(self, drive_id: str, folder_item: DriveItem, local_path: Path):
        """
        Recursively download a folder's contents using Microsoft Graph API.
        """
        if not local_path.exists():
            local_path.mkdir(parents=True, exist_ok=True)

        # Get children
        children = await self.graph_client.drives.by_drive_id(drive_id)\
            .items.by_drive_item_id(folder_item.id).children.get()

        if children and children.value:
            for item in children.value:
                item_path = local_path / item.name

                if item.folder:
                    # Recursively download subfolder
                    await self._download_folder_recursive(drive_id, item, item_path)
                else:
                    # Download file
                    await self.file_download(item.id, item_path)

    async def file_delete(self, item_id: str) -> bool:
        """
        Delete a file or folder in OneDrive by item ID using Microsoft Graph API.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Get item info for logging
            try:
                item = await self.graph_client.drives.by_drive_id(drive_id)\
                    .items.by_drive_item_id(item_id).get()
                item_name = item.name
                item_type = "folder" if item.folder else "file"
            except Exception:
                item_name = f"item {item_id}"
                item_type = "item"

            # Delete the item
            await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(item_id).delete()

            self.logger.info(f"Deleted {item_type}: {item_name}")
            return True

        except Exception as err:
            self.logger.error(f"Error deleting item {item_id}: {err}")
            raise RuntimeError(f"Error deleting item {item_id}: {err}") from err

    async def upload_file(self, file_path: Path, destination_folder: str = None) -> dict:
        """
        Upload a single file to OneDrive using Microsoft Graph API.
        """
        try:
            file_path = Path(file_path)
            if not file_path.exists():
                raise RuntimeError(f"File not found: {file_path}")

            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Ensure destination folder exists
            if destination_folder:
                folder_info = await self._ensure_folder(destination_folder, create=True)
                parent_id = folder_info.id
            else:
                root = await self.graph_client.drives.by_drive_id(drive_id).root.get()
                parent_id = root.id

            file_size = file_path.stat().st_size
            target_name = file_path.name

            self.logger.info(f"Uploading {target_name} ({file_size:,} bytes)")

            if file_size <= self.small_file_threshold:
                # Small file upload
                result = await self._upload_small_file(drive_id, parent_id, file_path, target_name)
            else:
                # Large file upload
                upload_session = await self._create_upload_session(drive_id, parent_id, target_name)
                result = await self._upload_large_file(upload_session, file_path)

            self.logger.info(f"Uploaded successfully: {result.name}")

            return {
                "name": result.name,
                "id": result.id,
                "webUrl": result.web_url,
                "size": getattr(result, 'size', file_size)
            }

        except Exception as err:
            self.logger.error(f"Error uploading file {file_path}: {err}")
            raise RuntimeError(f"Error uploading file {file_path}: {err}") from err

    async def upload_files(self, files: List[Path], destination_folder: str = None) -> List[dict]:
        """
        Upload multiple files to OneDrive using Microsoft Graph API.
        """
        uploaded_files = []

        for file_path in files:
            try:
                uploaded_item = await self.upload_file(file_path, destination_folder)
                uploaded_files.append(uploaded_item)
            except Exception as e:
                self.logger.error(f"Failed to upload {file_path}: {e}")
                continue

        return uploaded_files

    async def upload_folder(self, local_folder: Path, destination_folder: str = None) -> List[dict]:
        """
        Upload a local folder and its contents to OneDrive using Microsoft Graph API.
        """
        try:
            local_path = Path(local_folder)
            if not local_path.exists() or not local_path.is_dir():
                raise FileNotFoundError(
                    f"Local folder does not exist or is not a directory: {local_folder}"
                )

            uploaded_items = []

            # Get all files in the folder recursively
            for root, dirs, files in os.walk(local_path):
                relative_path = Path(root).relative_to(local_path)

                # Calculate OneDrive destination path
                if destination_folder:
                    onedrive_path = f"{destination_folder}/{relative_path}".replace("\\", "/").strip("/")
                else:
                    onedrive_path = str(relative_path).replace("\\", "/")

                # Ensure the directory exists in OneDrive
                if onedrive_path and onedrive_path != ".":
                    await self._ensure_folder(onedrive_path, create=True)

                # Upload files in this directory
                for file_name in files:
                    file_path = Path(root) / file_name
                    try:
                        uploaded_item = await self.upload_file(
                            file_path,
                            onedrive_path if onedrive_path != "." else None
                        )
                        uploaded_items.append(uploaded_item)
                    except Exception as e:
                        self.logger.error(f"Failed to upload {file_path}: {e}")
                        continue

            return uploaded_items

        except Exception as err:
            self.logger.error(f"Error uploading folder {local_folder}: {err}")
            raise RuntimeError(f"Error uploading folder {local_folder}: {err}") from err

    async def download_excel_file(self, item_id: str, destination: Path = None, as_pandas: bool = False):
        """
        Download an Excel file from OneDrive by item ID using Microsoft Graph API.
        If `as_pandas` is True, return as a pandas DataFrame.
        If `as_pandas` is False, save to the destination path.
        """
        try:
            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Get item info
            item = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(item_id).get()

            if not item.file:
                raise RuntimeError(f"Item {item_id} is not a file")

            # Get file content
            content = await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(item_id).content.get()

            if as_pandas:
                bytes_buffer = BytesIO(content)
                return pd.read_excel(bytes_buffer)
            else:
                if not destination:
                    raise ValueError("Destination path must be provided when `as_pandas` is False.")

                destination.parent.mkdir(parents=True, exist_ok=True)
                async with aiofiles.open(destination, "wb") as f:
                    await f.write(content)
                return str(destination)

        except Exception as err:
            self.logger.error(f"Error downloading Excel file {item_id}: {err}")
            raise RuntimeError(f"Error downloading Excel file {item_id}: {err}") from err

    async def upload_dataframe_as_excel(self, df: pd.DataFrame, file_name: str, destination_folder: str = None) -> dict:
        """
        Upload a pandas DataFrame as an Excel file to OneDrive using Microsoft Graph API.
        """
        try:
            # Convert DataFrame to Excel bytes
            output = BytesIO()
            df.to_excel(output, index=False)
            excel_content = output.getvalue()

            drive_info = await self._resolve_drive()
            drive_id = drive_info.id

            # Ensure destination folder exists
            if destination_folder:
                folder_info = await self._ensure_folder(destination_folder, create=True)
                parent_id = folder_info.id
            else:
                root = await self.graph_client.drives.by_drive_id(drive_id).root.get()
                parent_id = root.id

            self.logger.info(f"Uploading DataFrame as Excel: {file_name}")

            # Upload the Excel content
            if len(excel_content) <= self.small_file_threshold:
                # Small file upload
                encoded_name = quote(file_name)
                request_path = f"{parent_id}:/{encoded_name}:"

                uploaded_item = await self.graph_client.drives.by_drive_id(drive_id)\
                    .items.by_drive_item_id(request_path).content.put(excel_content)
            else:
                # Large file upload (create upload session)
                upload_session = await self._create_upload_session(drive_id, parent_id, file_name)
                uploaded_item = await self._upload_large_file_content(upload_session, excel_content, file_name)

            self.logger.info(f"Uploaded DataFrame as Excel successfully: {uploaded_item.name}")

            return {
                "name": uploaded_item.name,
                "id": uploaded_item.id,
                "webUrl": uploaded_item.web_url
            }

        except Exception as err:
            self.logger.error(f"Error uploading DataFrame as Excel file {file_name}: {err}")
            raise RuntimeError(f"Error uploading DataFrame as Excel file {file_name}: {err}") from err

    # Helper methods (similar to SharepointClient)

    async def _upload_small_file(self, drive_id: str, parent_id: str, local_path: Path, target_name: str) -> DriveItem:
        """Upload small file using Graph API."""
        try:
            async with aiofiles.open(local_path, "rb") as f:
                content = await f.read()

            # URL encode the target name to handle special characters
            encoded_name = quote(target_name)
            request_path = f"{parent_id}:/{encoded_name}:"

            return await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(request_path).content.put(content)

        except Exception as e:
            raise RuntimeError(f"Small file upload failed for {target_name}: {e}") from e

    async def _create_upload_session(self, drive_id: str, parent_id: str, target_name: str) -> UploadSession:
        """Create upload session for large files using Graph API."""
        try:
            body = CreateUploadSessionPostRequestBody()
            body.item = DriveItemUploadableProperties()
            body.item.additional_data = {"@microsoft.graph.conflictBehavior": "replace"}

            # URL encode the target name to handle special characters
            encoded_name = quote(target_name)

            return await self.graph_client.drives.by_drive_id(drive_id)\
                .items.by_drive_item_id(f"{parent_id}:/{encoded_name}:/")\
                .create_upload_session.post(body)

        except Exception as e:
            raise RuntimeError(f"Upload session creation failed for {target_name}: {e}") from e

    async def _upload_large_file(self, upload_session: UploadSession, local_path: Union[str, Path]) -> DriveItem:
        """Upload large file using resumable upload session."""
        file_size = os.path.getsize(local_path)
        uploaded = 0

        async with aiohttp.ClientSession() as session:
            async with aiofiles.open(local_path, "rb") as f:
                with tqdm(total=file_size, unit='B', unit_scale=True, desc=f'Uploading {Path(local_path).name}') as pbar:  # noqa
                    while uploaded < file_size:
                        chunk = await f.read(self.chunk_size)
                        if not chunk:
                            break

                        start = uploaded
                        end = uploaded + len(chunk) - 1

                        headers = {
                            "Content-Length": str(len(chunk)),
                            "Content-Range": f"bytes {start}-{end}/{file_size}"
                        }

                        async with session.put(
                            upload_session.upload_url,
                            headers=headers,
                            data=chunk
                        ) as response:
                            if response.status in (200, 201):
                                # Upload complete
                                pbar.update(file_size - uploaded)
                                result_data = await response.json()

                                # Convert to DriveItem (simplified)
                                drive_item = DriveItem()
                                drive_item.name = result_data.get('name')
                                drive_item.size = result_data.get('size')
                                drive_item.web_url = result_data.get('webUrl')
                                drive_item.additional_data = result_data

                                return drive_item

                            elif response.status == 202:
                                # Continue uploading
                                uploaded = end + 1
                                pbar.update(len(chunk))

                                # Check for retry-after header
                                if (retry_after := response.headers.get('Retry-After')):
                                    await asyncio.sleep(int(retry_after))
                                continue

                            else:
                                error_text = await response.text()
                                raise RuntimeError(f"Chunk upload failed: {response.status} {error_text}")

        raise RuntimeError("Upload session completed without final item response")

    async def _upload_large_file_content(
        self,
        upload_session: UploadSession,
        content: bytes,
        file_name: str
    ) -> DriveItem:
        """Upload large content using resumable upload session."""
        file_size = len(content)
        uploaded = 0

        async with aiohttp.ClientSession() as session:
            with tqdm(total=file_size, unit='B', unit_scale=True, desc=f'Uploading {file_name}') as pbar:
                while uploaded < file_size:
                    chunk = content[uploaded:uploaded + self.chunk_size]
                    if not chunk:
                        break

                    start = uploaded
                    end = uploaded + len(chunk) - 1

                    headers = {
                        "Content-Length": str(len(chunk)),
                        "Content-Range": f"bytes {start}-{end}/{file_size}"
                    }

                    async with session.put(
                        upload_session.upload_url,
                        headers=headers,
                        data=chunk
                    ) as response:
                        if response.status in (200, 201):
                            # Upload complete
                            pbar.update(file_size - uploaded)
                            result_data = await response.json()

                            # Convert to DriveItem (simplified)
                            drive_item = DriveItem()
                            drive_item.name = result_data.get('name')
                            drive_item.size = result_data.get('size')
                            drive_item.web_url = result_data.get('webUrl')
                            drive_item.additional_data = result_data

                            return drive_item

                        elif response.status == 202:
                            # Continue uploading
                            uploaded = end + 1
                            pbar.update(len(chunk))

                            # Check for retry-after header
                            if (retry_after := response.headers.get('Retry-After')):
                                await asyncio.sleep(int(retry_after))
                            continue

                        else:
                            error_text = await response.text()
                            raise RuntimeError(f"Chunk upload failed: {response.status} {error_text}")

        raise RuntimeError("Upload session completed without final item response")

    def _get_item_path_from_item(self, item) -> str:
        """
        Extract the full path from a DriveItem object.
        """
        try:
            # Try to get path from parent_reference
            if hasattr(item, 'parent_reference') and item.parent_reference and item.parent_reference.path:
                parent_path = item.parent_reference.path or ""

                # Clean up the parent path
                if parent_path.startswith("/drive/root:"):
                    parent_path = parent_path[12:]
                elif parent_path.startswith("/drives/") and "/root:" in parent_path:
                    parent_path = parent_path.split("/root:")[-1]

                # Build the full path
                if parent_path:
                    full_path = f"{parent_path}/{item.name}".replace("//", "/").lstrip("/")
                else:
                    full_path = item.name or ""

                return full_path

            # Fallback: try to get path from web_url if available
            if hasattr(item, 'web_url') and item.web_url:
                try:
                    # Extract path from OneDrive web URL
                    web_url = item.web_url
                    if "/personal/" in web_url and "/_layouts/15/onedrive.aspx" in web_url:
                        # Personal OneDrive URL format
                        return item.name or ""
                    else:
                        return unquote(item.name or "")
                except Exception:
                    pass

            # Final fallback: just return the filename
            return item.name or ""

        except Exception as e:
            self.logger.debug(f"Error extracting path from item: {e}")
            return item.name or ""

    async def test_permissions(self) -> Dict[str, Any]:
        """
        Test OneDrive permissions using Microsoft Graph API.
        """
        results = {
            "drive_access": False,
            "folder_access": False,
            "upload_access": False,
            "errors": []
        }

        try:
            # Test 1: Drive access
            drive_info = await self._resolve_drive()
            results["drive_access"] = True
            self.logger.info(f"Drive access: {drive_info.name or 'Personal OneDrive'}")

            # Test 2: Folder access (list root)
            file_list = await self.file_list()
            results["folder_access"] = True
            self.logger.info(f"Folder access: Listed {len(file_list)} items")

            # Test 3: Folder creation (upload capability test)
            test_folder = await self._ensure_folder("test-folder-permissions", create=True)
            results["upload_access"] = True
            self.logger.info("Upload permissions confirmed")

            # Clean up test folder
            with contextlib.suppress(Exception):
                await self.file_delete(test_folder.id)
                self.logger.info("Test folder cleaned up")

        except Exception as e:
            results["errors"].append(str(e))
            self.logger.error(f"Permission test failed: {e}")

        return results

    async def close(self):
        """Clean up resources."""
        await super().close()
        self._drive_info = None
        self._drive_id = None



from typing import Dict, List, Optional, Any
import asyncio
import json
import aiofiles
from pathlib import Path
from .abstract import ConversationMemory, ConversationHistory, ConversationTurn


class FileConversationMemory(ConversationMemory):
    """File-based implementation of conversation memory."""

    def __init__(self, base_path: str = "./conversations"):
        self.base_path = Path(base_path)
        self.base_path.mkdir(exist_ok=True)
        self._lock = asyncio.Lock()

    def _get_file_path(
        self,
        user_id: str,
        session_id: str,
        chatbot_id: Optional[str] = None
    ) -> Path:
        """Get file path for a conversation history."""
        user_dir = self.base_path / str(user_id)
        if chatbot_id:
            user_dir = user_dir / str(chatbot_id)
        user_dir.mkdir(parents=True, exist_ok=True)
        return user_dir / f"{session_id}.json"

    async def create_history(
        self,
        user_id: str,
        session_id: str,
        metadata: Optional[Dict[str, Any]] = None,
        chatbot_id: Optional[str] = None
    ) -> ConversationHistory:
        """Create a new conversation history."""
        async with self._lock:
            history = ConversationHistory(
                session_id=session_id,
                user_id=user_id,
                chatbot_id=chatbot_id,
                metadata=metadata or {}
            )

            file_path = self._get_file_path(user_id, session_id, chatbot_id)
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(history.to_dict(), f, indent=2, ensure_ascii=False, default=str)

            return history

    async def get_history(
        self,
        user_id: str,
        session_id: str,
        chatbot_id: Optional[str] = None
    ) -> Optional[ConversationHistory]:
        """Get a conversation history."""
        async with self._lock:
            file_path = self._get_file_path(user_id, session_id, chatbot_id)
            if not file_path.exists():
                return None

            try:
                async with aiofiles.open(file_path, 'r', encoding='utf-8') as f:
                    content = await f.read()
                data = json.loads(content)
                return ConversationHistory.from_dict(data)
            except (TypeError, KeyError, ValueError):
                return None

    async def update_history(self, history: ConversationHistory) -> None:
        """Update a conversation history."""
        async with self._lock:
            file_path = self._get_file_path(
                history.user_id,
                history.session_id,
                history.chatbot_id
            )
            async with aiofiles.open(file_path, 'w', encoding='utf-8') as f:
                await f.write(json.dumps(history.to_dict(), indent=2, ensure_ascii=False, default=str))

    async def add_turn(
        self,
        user_id: str,
        session_id: str,
        turn: ConversationTurn,
        chatbot_id: Optional[str] = None
    ) -> None:
        """Add a turn to the conversation."""
        history = await self.get_history(user_id, session_id, chatbot_id)
        if history:
            history.add_turn(turn)
            await self.update_history(history)

    async def clear_history(
        self,
        user_id: str,
        session_id: str,
        chatbot_id: Optional[str] = None
    ) -> None:
        """Clear a conversation history."""
        history = await self.get_history(user_id, session_id, chatbot_id)
        if history:
            history.clear_turns()
            await self.update_history(history)

    async def list_sessions(
        self,
        user_id: str,
        chatbot_id: Optional[str] = None
    ) -> List[str]:
        """List all session IDs for a user."""
        async with self._lock:
            base_user_dir = self.base_path / str(user_id)
            if not base_user_dir.exists():
                return []

            sessions: List[str] = []
            seen = set()
            if chatbot_id is None:
                for file_path in base_user_dir.glob("*.json"):
                    if file_path.stem not in seen:
                        seen.add(file_path.stem)
                        sessions.append(file_path.stem)
                for subdir in base_user_dir.iterdir():
                    if subdir.is_dir():
                        for file_path in subdir.glob("*.json"):
                            if file_path.stem not in seen:
                                seen.add(file_path.stem)
                                sessions.append(file_path.stem)
            else:
                chatbot_dir = base_user_dir / str(chatbot_id)
                if chatbot_dir.exists():
                    for file_path in chatbot_dir.glob("*.json"):
                        if file_path.stem not in seen:
                            seen.add(file_path.stem)
                            sessions.append(file_path.stem)

            return sessions

    async def delete_history(
        self,
        user_id: str,
        session_id: str,
        chatbot_id: Optional[str] = None
    ) -> bool:
        """Delete a conversation history entirely."""
        async with self._lock:
            file_path = self._get_file_path(user_id, session_id, chatbot_id)
            if file_path.exists():
                file_path.unlink()
                return True
            return False



"""TelegramCrewTransport — top-level orchestrator for multi-agent crew.

Manages the full lifecycle of a multi-agent crew in a Telegram supergroup:
coordinator bot startup, agent wrapper creation, agent registration,
aiogram polling, and graceful shutdown.

Usage::

    config = TelegramCrewConfig.from_yaml("crew.yaml")
    async with TelegramCrewTransport(config) as transport:
        # transport is running — agents respond to @mentions
        await asyncio.Event().wait()  # or your application loop
"""
import asyncio
import logging
from datetime import datetime, timezone
from typing import Dict, List, Optional, TYPE_CHECKING

from aiogram import Bot, Dispatcher
from aiogram.client.default import DefaultBotProperties
from aiogram.enums import ParseMode

from .agent_card import AgentCard, AgentSkill
from .config import TelegramCrewConfig, CrewAgentEntry
from .coordinator import CoordinatorBot
from .crew_wrapper import CrewAgentWrapper
from .mention import format_reply
from .payload import DataPayload
from .registry import CrewRegistry

if TYPE_CHECKING:
    from ....bots.abstract import AbstractBot

logger = logging.getLogger(__name__)

# Rate limit delay between consecutive bot startups (seconds)
_STARTUP_DELAY = 0.3


class TelegramCrewTransport:
    """Orchestrator for a multi-agent crew in a Telegram supergroup.

    Manages the lifecycle of all bots: a coordinator bot and one aiogram
    ``Bot`` + ``CrewAgentWrapper`` per configured agent.

    Args:
        config: The ``TelegramCrewConfig`` describing the crew setup.
        bot_manager: Optional ``BotManager`` instance for retrieving agents
            by ``chatbot_id``.  When provided, ``start()`` will look up
            agent instances automatically.
    """

    def __init__(
        self,
        config: TelegramCrewConfig,
        bot_manager: Optional[object] = None,
    ) -> None:
        self.config = config
        self.bot_manager = bot_manager
        self.registry = CrewRegistry()
        self.coordinator: Optional[CoordinatorBot] = None
        self._wrappers: Dict[str, CrewAgentWrapper] = {}
        self._bots: Dict[str, Bot] = {}
        self._dispatchers: Dict[str, Dispatcher] = {}
        self._polling_tasks: List[asyncio.Task] = []
        self._payload: Optional[DataPayload] = None
        self.logger = logging.getLogger(__name__)

    # ------------------------------------------------------------------
    # Factory
    # ------------------------------------------------------------------

    @classmethod
    def from_config(
        cls,
        config: TelegramCrewConfig,
        bot_manager: Optional[object] = None,
    ) -> "TelegramCrewTransport":
        """Construct a transport from a ``TelegramCrewConfig``.

        Args:
            config: Validated crew configuration.
            bot_manager: Optional ``BotManager`` for agent retrieval.

        Returns:
            A new ``TelegramCrewTransport`` instance (not yet started).
        """
        return cls(config=config, bot_manager=bot_manager)

    # ------------------------------------------------------------------
    # Async context manager
    # ------------------------------------------------------------------

    async def __aenter__(self) -> "TelegramCrewTransport":
        await self.start()
        return self

    async def __aexit__(self, *exc) -> None:
        await self.stop()

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    async def start(self) -> None:
        """Start the crew transport.

        1. Create a shared ``DataPayload`` for file exchange.
        2. Start the ``CoordinatorBot`` (sends + pins the registry message).
        3. For each agent in config, create a ``Bot``, ``Dispatcher``,
           ``CrewAgentWrapper``, build an ``AgentCard``, register with the
           coordinator, and start aiogram polling.
        """
        self.logger.info(
            "Starting TelegramCrewTransport for group %d with %d agents",
            self.config.group_id,
            len(self.config.agents),
        )

        # Shared DataPayload for file exchange
        self._payload = DataPayload(
            temp_dir=self.config.temp_dir,
            max_file_size_mb=self.config.max_file_size_mb,
            allowed_mime_types=self.config.allowed_mime_types,
        )

        # Start coordinator bot
        self.coordinator = CoordinatorBot(
            token=self.config.coordinator_token,
            group_id=self.config.group_id,
            registry=self.registry,
            username=self.config.coordinator_username,
        )
        await self.coordinator.start()
        self.logger.info("CoordinatorBot started")

        # Start each agent
        for agent_name, entry in self.config.agents.items():
            try:
                await self._start_agent(agent_name, entry)
                await asyncio.sleep(_STARTUP_DELAY)
            except Exception as e:
                self.logger.error(
                    "Failed to start agent %s: %s", agent_name, e, exc_info=True
                )

        self.logger.info(
            "TelegramCrewTransport started: %d agents online",
            len(self._wrappers),
        )

    async def _start_agent(
        self, agent_name: str, entry: CrewAgentEntry
    ) -> None:
        """Start a single agent bot, wrapper, and polling task.

        Args:
            agent_name: Human-readable agent name (from config dict key).
            entry: The ``CrewAgentEntry`` config for this agent.
        """
        # Retrieve the AI agent instance
        agent = await self._get_agent(agent_name, entry)

        # Create aiogram Bot
        bot = Bot(
            token=entry.bot_token,
            default=DefaultBotProperties(parse_mode=ParseMode.MARKDOWN),
        )
        self._bots[entry.username] = bot

        # Get bot user info for the AgentCard
        bot_user = await bot.me()

        # Build AgentCard
        skills = [
            AgentSkill(name=s.get("name", ""), description=s.get("description", ""))
            for s in entry.skills
        ]
        now = datetime.now(timezone.utc)
        card = AgentCard(
            agent_id=entry.chatbot_id,
            agent_name=agent_name,
            telegram_username=entry.username,
            telegram_user_id=bot_user.id,
            model=getattr(agent, "model", "unknown") if agent else "unknown",
            skills=skills,
            tags=entry.tags,
            accepts_files=entry.accepts_files,
            emits_files=entry.emits_files,
            joined_at=now,
            last_seen=now,
        )

        # Create wrapper
        wrapper = CrewAgentWrapper(
            bot=bot,
            agent=agent,
            card=card,
            group_id=self.config.group_id,
            coordinator=self.coordinator,
            payload=self._payload,
        )
        self._wrappers[entry.username] = wrapper

        # Register with coordinator
        await self.coordinator.on_agent_join(card)

        # Create dispatcher and include wrapper router
        dp = Dispatcher()
        dp.include_router(wrapper.router)
        self._dispatchers[entry.username] = dp

        # Start polling in background
        task = asyncio.create_task(
            self._run_polling(entry.username, bot, dp),
            name=f"polling-{entry.username}",
        )
        self._polling_tasks.append(task)

        self.logger.info(
            "Agent @%s (%s) started and registered",
            entry.username, agent_name,
        )

    async def _get_agent(
        self, agent_name: str, entry: CrewAgentEntry
    ) -> Optional["AbstractBot"]:
        """Retrieve an agent instance from the BotManager.

        Args:
            agent_name: Display name of the agent.
            entry: The ``CrewAgentEntry`` config.

        Returns:
            The agent instance, or ``None`` if not available.
        """
        if self.bot_manager is None:
            self.logger.warning(
                "No BotManager provided — agent %s will have a stub agent",
                agent_name,
            )
            return None

        try:
            agent = await self.bot_manager.get_bot(entry.chatbot_id)
        except Exception as e:
            self.logger.error(
                "Failed to get agent %s from BotManager: %s",
                entry.chatbot_id, e,
            )
            return None

        if not agent:
            self.logger.error(
                "Agent '%s' (chatbot_id=%s) not found in BotManager",
                agent_name, entry.chatbot_id,
            )
            return None

        # Apply system prompt override if specified
        if entry.system_prompt_override and hasattr(agent, "system_prompt"):
            agent.system_prompt = entry.system_prompt_override

        return agent

    async def _run_polling(
        self, username: str, bot: Bot, dp: Dispatcher
    ) -> None:
        """Run aiogram polling for a single bot (background task).

        Args:
            username: Telegram username of the bot.
            bot: The aiogram ``Bot`` instance.
            dp: The aiogram ``Dispatcher``.
        """
        try:
            self.logger.info("Starting polling for @%s", username)
            await dp.start_polling(bot)
        except asyncio.CancelledError:
            self.logger.info("Polling cancelled for @%s", username)
        except Exception as e:
            self.logger.error(
                "Polling error for @%s: %s", username, e, exc_info=True
            )

    async def stop(self) -> None:
        """Stop the crew transport gracefully.

        1. Cancel all polling tasks.
        2. Unregister all agents from the coordinator.
        3. Stop the coordinator bot.
        4. Close all bot sessions.
        5. Clean up payload temp files.
        """
        self.logger.info("Stopping TelegramCrewTransport")

        # Cancel polling tasks
        for task in self._polling_tasks:
            task.cancel()
        if self._polling_tasks:
            await asyncio.gather(*self._polling_tasks, return_exceptions=True)
        self._polling_tasks.clear()

        # Stop dispatchers
        for username, dp in self._dispatchers.items():
            try:
                await dp.stop_polling()
            except Exception:
                pass

        # Unregister agents from coordinator
        if self.coordinator:
            for username in list(self._wrappers.keys()):
                try:
                    await self.coordinator.on_agent_leave(username)
                except Exception as e:
                    self.logger.warning(
                        "Error unregistering @%s: %s", username, e
                    )

            # Stop coordinator
            try:
                await self.coordinator.stop()
            except Exception as e:
                self.logger.warning("Error stopping coordinator: %s", e)

        # Close all bot sessions
        for username, bot in self._bots.items():
            try:
                await bot.session.close()
            except Exception as e:
                self.logger.warning(
                    "Error closing session for @%s: %s", username, e
                )

        # Clean up temp files
        if self._payload:
            self._payload.cleanup_all()

        self._wrappers.clear()
        self._bots.clear()
        self._dispatchers.clear()

        self.logger.info("TelegramCrewTransport stopped")

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    async def send_message(
        self,
        from_username: str,
        mention: str,
        text: str,
        reply_to_message_id: Optional[int] = None,
    ) -> None:
        """Send a text message from a specific agent bot.

        Args:
            from_username: Telegram username of the sending agent.
            mention: @mention to include in the message.
            text: The message text.
            reply_to_message_id: Optional message ID to reply to.

        Raises:
            KeyError: If ``from_username`` is not a registered agent.
        """
        bot = self._get_bot(from_username)
        full_text = format_reply(mention, text)
        await bot.send_message(
            chat_id=self.config.group_id,
            text=full_text,
            reply_to_message_id=reply_to_message_id,
        )

    async def send_document(
        self,
        from_username: str,
        mention: str,
        file_path: str,
        caption: str = "",
        reply_to_message_id: Optional[int] = None,
    ) -> None:
        """Send a document from a specific agent bot.

        Args:
            from_username: Telegram username of the sending agent.
            mention: @mention to include in the caption.
            file_path: Path to the file to send.
            caption: Optional caption text.
            reply_to_message_id: Optional message ID to reply to.

        Raises:
            KeyError: If ``from_username`` is not a registered agent.
        """
        if self._payload is None:
            raise RuntimeError("DataPayload not initialized — call start() first")

        bot = self._get_bot(from_username)
        full_caption = format_reply(mention, caption) if caption else mention
        await self._payload.send_document(
            bot=bot,
            chat_id=self.config.group_id,
            file_path=file_path,
            caption=full_caption,
            reply_to_message_id=reply_to_message_id,
        )

    def list_online_agents(self) -> List[AgentCard]:
        """Return a list of currently active (non-offline) agents.

        Returns:
            List of ``AgentCard`` instances for agents that are online.
        """
        return self.registry.list_active()

    def get_wrapper(self, username: str) -> Optional[CrewAgentWrapper]:
        """Get the ``CrewAgentWrapper`` for a specific agent.

        Args:
            username: Telegram username of the agent (without @).

        Returns:
            The wrapper instance, or ``None`` if not found.
        """
        return self._wrappers.get(username.lstrip("@"))

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _get_bot(self, username: str) -> Bot:
        """Get the aiogram Bot for a given username.

        Args:
            username: Telegram username of the agent.

        Returns:
            The aiogram ``Bot`` instance.

        Raises:
            KeyError: If the username is not registered.
        """
        clean = username.lstrip("@")
        if clean not in self._bots:
            raise KeyError(
                f"No bot registered for @{clean}. "
                f"Available: {list(self._bots.keys())}"
            )
        return self._bots[clean]



"""Simple in-memory storage for agent question/answer pairs keyed by turn_id."""

import asyncio
from typing import Any, Dict, Optional


class AnswerMemory:
    """Store and retrieve question/answer interactions by turn identifier."""

    def __init__(self, agent_id: str) -> None:
        self.agent_id = agent_id
        self._interactions: Dict[str, Dict[str, Any]] = {self.agent_id: {}}
        self._lock = asyncio.Lock()

    async def store_interaction(self, turn_id: str, question: str, answer: Any) -> None:
        """Persist a question/answer pair under the provided turn identifier."""
        if not turn_id:
            raise ValueError("turn_id is required to store an interaction")

        async with self._lock:
            self._interactions[self.agent_id][turn_id] = {
                "question": question,
                "answer": answer,
            }

    async def get(self, turn_id: str) -> Optional[Dict[str, Any]]:
        """Retrieve a stored interaction by turn identifier."""
        if not turn_id:
            return None

        async with self._lock:
            return self._interactions.get(self.agent_id, {}).get(turn_id)


# Backward-compatible alias
AgentMemory = AnswerMemory



"""REST handler for chat interaction persistence.

Provides endpoints to list, load, and delete chat conversations
stored via ChatStorage (Redis + DocumentDB).
"""

from typing import Optional
from datetime import datetime

from aiohttp import web
from navigator.views import BaseView
from navigator_session import get_session
from navigator_auth.decorators import is_authenticated, user_session
from navigator_auth.conf import AUTH_SESSION_OBJECT


@is_authenticated()
@user_session()
class ChatInteractionHandler(BaseView):
    """Manage persisted chat interactions.

    GET    /api/v1/chat/interactions          — list conversations
    GET    /api/v1/chat/interactions/{sid}     — load messages for a session
    POST   /api/v1/chat/interactions          — create a conversation
    PUT    /api/v1/chat/interactions/{sid}     — update conversation title
    DELETE /api/v1/chat/interactions/{sid}     — delete a conversation
    """

    _logger_name: str = "Parrot.ChatInteraction"

    def _get_storage(self):
        """Retrieve ChatStorage from the application context."""
        try:
            return self.request.app["chat_storage"]
        except KeyError:
            self.logger.error("chat_storage not found in app context")
            return None

    async def _get_user_id(self) -> Optional[str]:
        """Extract user_id from the authenticated session."""
        # 1. Try the user object set by @user_session / middleware
        user = getattr(self.request, "user", None)
        if user:
            uid = getattr(user, "user_id", None) or getattr(user, "id", None)
            if uid:
                return str(uid)
        # 2. Extract from session stored in Redis
        try:
            session = await get_session(self.request)
        except Exception:
            return None
        if session:
            userinfo = session.get(AUTH_SESSION_OBJECT, {})
            if isinstance(userinfo, dict):
                user_id = userinfo.get("user_id")
                if user_id:
                    return str(user_id)
            user_id = session.get("user_id")
            if user_id:
                return str(user_id)
        return None

    async def get(self) -> web.Response:
        """List conversations or load a specific session's messages."""
        storage = self._get_storage()
        if storage is None:
            self.error(
                response={"message": "Chat storage not available"},
                status=503,
            )

        user_id = await self._get_user_id()
        if not user_id:
            self.error(
                response={"message": "User ID not found in session"},
                status=401,
            )

        # Check for session_id in URL path
        session_id = self.request.match_info.get("session_id")

        if session_id:
            return await self._get_conversation_messages(storage, user_id, session_id)
        return await self._list_conversations(storage, user_id)

    async def _list_conversations(
        self, storage, user_id: str
    ) -> web.Response:
        """List all conversations for the authenticated user."""
        qs = self.get_arguments(self.request)
        agent_id = qs.get("agent_id")
        limit = int(qs.get("limit", 50))
        since_str = qs.get("since")

        since: Optional[datetime] = None
        if since_str:
            try:
                since = datetime.fromisoformat(since_str)
            except ValueError:
                self.error(
                    response={"message": "Invalid 'since' format, use ISO 8601"},
                    status=400,
                )

        conversations = await storage.list_user_conversations(
            user_id=user_id,
            agent_id=agent_id,
            limit=limit,
            since=since,
        )
        return self.json_response({
            "conversations": conversations,
            "count": len(conversations),
        })

    async def _get_conversation_messages(
        self, storage, user_id: str, session_id: str
    ) -> web.Response:
        """Load messages for a specific conversation."""
        qs = self.get_arguments(self.request)
        agent_id = qs.get("agent_id")
        limit = int(qs.get("limit", 50))

        messages = await storage.load_conversation(
            user_id=user_id,
            session_id=session_id,
            agent_id=agent_id,
            limit=limit,
        )

        metadata = await storage.get_conversation_metadata(session_id)

        return self.json_response({
            "session_id": session_id,
            "messages": messages,
            "count": len(messages),
            "metadata": metadata,
        })

    async def post(self) -> web.Response:
        """Create a new conversation."""
        storage = self._get_storage()
        if storage is None:
            self.error(
                response={"message": "Chat storage not available"},
                status=503,
            )

        user_id = await self._get_user_id()
        if not user_id:
            self.error(
                response={"message": "User ID not found in session"},
                status=401,
            )

        try:
            body = await self.json_data()
        except Exception:
            body = {}

        session_id = body.get("session_id")
        agent_id = body.get("agent_id")
        title = body.get("title", "New Conversation")

        if not session_id or not agent_id:
            self.error(
                response={"message": "session_id and agent_id are required"},
                status=400,
            )

        result = await storage.create_conversation(
            user_id=user_id,
            session_id=session_id,
            agent_id=agent_id,
            title=title,
        )

        if result is not None:
            return self.json_response({
                "message": "Conversation created",
                "conversation": result,
            }, status=201)
        self.error(
            response={"message": "Failed to create conversation"},
            status=500,
        )

    async def put(self) -> web.Response:
        """Update the title of a conversation."""
        storage = self._get_storage()
        if storage is None:
            self.error(
                response={"message": "Chat storage not available"},
                status=503,
            )

        user_id = await self._get_user_id()
        if not user_id:
            self.error(
                response={"message": "User ID not found in session"},
                status=401,
            )

        session_id = self.request.match_info.get("session_id")
        if not session_id:
            self.error(
                response={"message": "session_id is required in path"},
                status=400,
            )

        try:
            body = await self.json_data()
        except Exception:
            body = {}

        title = body.get("title")
        if not title:
            self.error(
                response={"message": "title is required"},
                status=400,
            )

        updated = await storage.update_conversation_title(
            session_id=session_id,
            title=title,
        )

        if updated:
            return self.json_response({
                "message": "Title updated",
                "session_id": session_id,
                "title": title,
            })
        self.error(
            response={"message": f"Failed to update conversation {session_id}"},
            status=500,
        )

    async def delete(self) -> web.Response:
        """Delete a conversation by session_id."""
        storage = self._get_storage()
        if storage is None:
            self.error(
                response={"message": "Chat storage not available"},
                status=503,
            )

        user_id = await self._get_user_id()
        if not user_id:
            self.error(
                response={"message": "User ID not found in session"},
                status=401,
            )

        session_id = self.request.match_info.get("session_id")
        if not session_id:
            self.error(
                response={"message": "session_id is required in path"},
                status=400,
            )

        qs = self.get_arguments(self.request)
        agent_id = qs.get("agent_id")

        deleted = await storage.delete_conversation(
            user_id=user_id,
            session_id=session_id,
            agent_id=agent_id,
        )

        if deleted:
            return self.json_response({
                "message": f"Conversation {session_id} deleted",
                "session_id": session_id,
            })
        self.error(
            response={"message": f"Conversation {session_id} not found"},
            status=404,
        )

    async def patch(self) -> web.Response:
        """Delete a specific turn from a conversation."""
        storage = self._get_storage()
        if storage is None:
            self.error(
                response={"message": "Chat storage not available"},
                status=503,
            )

        user_id = await self._get_user_id()
        if not user_id:
            self.error(
                response={"message": "User ID not found in session"},
                status=401,
            )

        session_id = self.request.match_info.get("session_id")
        if not session_id:
            self.error(
                response={"message": "session_id is required in path"},
                status=400,
            )

        try:
            body = await self.json_data()
        except Exception:
            body = {}

        action = body.get("action")
        if action != "delete_turn":
            self.error(
                response={"message": f"Unknown action: {action}"},
                status=400,
            )

        turn_id = body.get("turn_id")
        if not turn_id:
            self.error(
                response={"message": "turn_id is required"},
                status=400,
            )

        deleted = await storage.delete_turn(
            session_id=session_id,
            turn_id=turn_id,
        )

        if deleted:
            return self.json_response({
                "message": f"Turn {turn_id} deleted",
                "session_id": session_id,
                "turn_id": turn_id,
            })
        self.error(
            response={"message": f"Turn {turn_id} not found"},
            status=404,
        )



"""Microsoft SQL Server database source for DatabaseToolkit.

Implements ``AbstractDatabaseSource`` for MSSQL using the asyncdb ``mssql`` driver.
Overrides ``validate_query()`` to allow EXEC/EXECUTE statements (stored procedures)
in addition to standard T-SQL. Includes stored procedures in metadata discovery.

Part of FEAT-062 — DatabaseToolkit.
"""
from __future__ import annotations

import logging
import re
import time
from typing import Any

from parrot.tools.database.base import (
    AbstractDatabaseSource,
    ColumnMeta,
    MetadataResult,
    QueryResult,
    RowResult,
    TableMeta,
    ValidationResult,
    _validate_sql_identifier,
)
from parrot.tools.database.sources import register_source

# Pattern to detect stored procedure execution statements
_EXEC_PATTERN = re.compile(r"^\s*(EXEC|EXECUTE)\s+", re.IGNORECASE)


@register_source("mssql")
class MSSQLSource(AbstractDatabaseSource):
    """Microsoft SQL Server database source with stored procedure support.

    Uses the asyncdb ``mssql`` driver. Validates queries with the ``tsql``
    sqlglot dialect, with special handling for EXEC/EXECUTE stored procedure calls.
    Exposes stored procedures alongside tables in metadata discovery.
    """

    driver = "mssql"
    sqlglot_dialect = "tsql"

    def __init__(self) -> None:
        """Initialize MSSQLSource with a logger."""
        self.logger = logging.getLogger("Parrot.Toolkits.Database.MSSQL")

    async def get_default_credentials(self) -> dict[str, Any]:
        """Return default MSSQL credentials.

        Returns:
            Empty dict (no default MSSQL credentials configured).
        """
        from parrot.interfaces.database import get_default_credentials
        dsn = get_default_credentials("mssql")
        return {"dsn": dsn} if dsn else {}

    async def validate_query(self, query: str) -> ValidationResult:
        """Validate a T-SQL query, including EXEC/EXECUTE statements.

        EXEC and EXECUTE are valid MSSQL stored procedure calls. The base
        sqlglot validation may reject them, so we pre-check for this pattern
        and return valid=True immediately for SP invocations.

        Standard SELECT/INSERT/UPDATE/DELETE queries are validated via the
        base class using the ``tsql`` sqlglot dialect.

        Args:
            query: T-SQL query string or EXEC/EXECUTE statement.

        Returns:
            ValidationResult indicating whether the query is valid.
        """
        if _EXEC_PATTERN.match(query):
            self.logger.debug("EXEC/EXECUTE statement detected — validated as valid")
            return ValidationResult(valid=True, dialect="tsql")
        # Delegate to base sqlglot validation with tsql dialect
        return await super().validate_query(query)

    async def get_metadata(
        self,
        credentials: dict[str, Any],
        tables: list[str] | None = None,
    ) -> MetadataResult:
        """Discover MSSQL schema, including tables and stored procedures.

        Queries both ``INFORMATION_SCHEMA.COLUMNS`` for table metadata and
        ``sys.procedures``/``sys.parameters`` for stored procedure metadata.
        Stored procedures are returned as TableMeta entries with
        ``schema_name="stored_procedures"``.

        Args:
            credentials: Connection credentials.
            tables: Optional list of specific table names to inspect.

        Returns:
            MetadataResult with table, column, and stored procedure metadata.
        """
        self.logger.debug("get_metadata called, tables=%s", tables)
        # Validate table names before acquiring a connection (fail-fast on injection)
        if tables:
            validated_tables = [_validate_sql_identifier(t, "table name") for t in tables]
        else:
            validated_tables = None

        dsn = credentials.get("dsn")
        params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("mssql", dsn, params)
        async with await db.connection() as conn:
            # Query table columns
            if validated_tables:
                placeholders = ", ".join([f"'{t}'" for t in validated_tables])
                col_sql = f"""
                    SELECT
                        c.TABLE_SCHEMA, c.TABLE_NAME, c.COLUMN_NAME,
                        c.DATA_TYPE, c.IS_NULLABLE, c.COLUMN_DEFAULT,
                        CASE WHEN kcu.COLUMN_NAME IS NOT NULL THEN 1 ELSE 0 END AS is_pk
                    FROM INFORMATION_SCHEMA.COLUMNS c
                    LEFT JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc
                        ON tc.TABLE_SCHEMA = c.TABLE_SCHEMA
                        AND tc.TABLE_NAME = c.TABLE_NAME
                        AND tc.CONSTRAINT_TYPE = 'PRIMARY KEY'
                    LEFT JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE kcu
                        ON kcu.CONSTRAINT_NAME = tc.CONSTRAINT_NAME
                        AND kcu.COLUMN_NAME = c.COLUMN_NAME
                    WHERE c.TABLE_NAME IN ({placeholders})
                    ORDER BY c.TABLE_NAME, c.ORDINAL_POSITION
                """
            else:
                col_sql = """
                    SELECT
                        c.TABLE_SCHEMA, c.TABLE_NAME, c.COLUMN_NAME,
                        c.DATA_TYPE, c.IS_NULLABLE, c.COLUMN_DEFAULT,
                        CASE WHEN kcu.COLUMN_NAME IS NOT NULL THEN 1 ELSE 0 END AS is_pk
                    FROM INFORMATION_SCHEMA.COLUMNS c
                    LEFT JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc
                        ON tc.TABLE_SCHEMA = c.TABLE_SCHEMA
                        AND tc.TABLE_NAME = c.TABLE_NAME
                        AND tc.CONSTRAINT_TYPE = 'PRIMARY KEY'
                    LEFT JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE kcu
                        ON kcu.CONSTRAINT_NAME = tc.CONSTRAINT_NAME
                        AND kcu.COLUMN_NAME = c.COLUMN_NAME
                    ORDER BY c.TABLE_NAME, c.ORDINAL_POSITION
                """
            col_rows = await conn.fetch_all(col_sql)

            # Query stored procedures and their parameters
            sproc_sql = """
                SELECT
                    p.name AS proc_name,
                    SCHEMA_NAME(p.schema_id) AS schema_name,
                    sp.name AS param_name,
                    TYPE_NAME(sp.user_type_id) AS param_type,
                    sp.is_output
                FROM sys.procedures p
                LEFT JOIN sys.parameters sp ON sp.object_id = p.object_id
                ORDER BY p.name, sp.parameter_id
            """
            try:
                sproc_rows = await conn.fetch_all(sproc_sql)
            except Exception:  # noqa: BLE001
                sproc_rows = []

        # Build table metadata
        tables_map: dict[str, TableMeta] = {}
        for row in (col_rows or []):
            row_dict = dict(row) if not isinstance(row, dict) else row
            table_name = row_dict.get("TABLE_NAME", row_dict.get("table_name", ""))
            if table_name not in tables_map:
                tables_map[table_name] = TableMeta(
                    name=table_name,
                    schema_name=row_dict.get("TABLE_SCHEMA", row_dict.get("table_schema")),
                )
            col = ColumnMeta(
                name=row_dict.get("COLUMN_NAME", row_dict.get("column_name", "")),
                data_type=row_dict.get("DATA_TYPE", row_dict.get("data_type", "unknown")),
                nullable=row_dict.get("IS_NULLABLE", row_dict.get("is_nullable", "YES")) == "YES",
                primary_key=bool(row_dict.get("is_pk", 0)),
                default=row_dict.get("COLUMN_DEFAULT", row_dict.get("column_default")),
            )
            tables_map[table_name].columns.append(col)

        # Build stored procedure metadata
        sproc_map: dict[str, TableMeta] = {}
        for row in (sproc_rows or []):
            row_dict = dict(row) if not isinstance(row, dict) else row
            proc_name = row_dict.get("proc_name", "")
            if not proc_name:
                continue
            if proc_name not in sproc_map:
                sproc_map[proc_name] = TableMeta(
                    name=proc_name,
                    schema_name="stored_procedures",
                )
            param_name = row_dict.get("param_name", "")
            if param_name:
                col = ColumnMeta(
                    name=param_name,
                    data_type=row_dict.get("param_type", "unknown"),
                    nullable=True,
                    primary_key=False,
                )
                sproc_map[proc_name].columns.append(col)

        all_tables = list(tables_map.values()) + list(sproc_map.values())
        return MetadataResult(driver=self.driver, tables=all_tables)

    async def query(
        self,
        credentials: dict[str, Any],
        sql: str,
        params: dict[str, Any] | None = None,
    ) -> QueryResult:
        """Execute a T-SQL query or stored procedure call.

        Supports both standard SELECT statements and EXEC/EXECUTE calls.

        Args:
            credentials: Connection credentials.
            sql: T-SQL query or EXEC statement.
            params: Optional query parameters.

        Returns:
            QueryResult with rows and execution metadata.
        """
        self.logger.debug("query called: %s", sql[:100])
        start = time.monotonic()
        dsn = credentials.get("dsn")
        conn_params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("mssql", dsn, conn_params)
        async with await db.connection() as conn:
            if params:
                if isinstance(params, dict):
                    rows = await conn.fetch_all(sql, **params)
                else:
                    rows = await conn.fetch_all(sql, *params)
            else:
                rows = await conn.fetch_all(sql)

        elapsed_ms = (time.monotonic() - start) * 1000
        rows_list = [dict(r) if not isinstance(r, dict) else r for r in (rows or [])]
        columns = list(rows_list[0].keys()) if rows_list else []

        return QueryResult(
            driver=self.driver,
            rows=rows_list,
            row_count=len(rows_list),
            columns=columns,
            execution_time_ms=round(elapsed_ms, 3),
        )

    async def query_row(
        self,
        credentials: dict[str, Any],
        sql: str,
        params: dict[str, Any] | None = None,
    ) -> RowResult:
        """Execute a T-SQL query and return the first row.

        Args:
            credentials: Connection credentials.
            sql: T-SQL query string.
            params: Optional query parameters.

        Returns:
            RowResult with a single row or found=False.
        """
        self.logger.debug("query_row called: %s", sql[:100])
        start = time.monotonic()
        dsn = credentials.get("dsn")
        conn_params = credentials.get("params", credentials if "host" in credentials else None)

        db = self._get_db("mssql", dsn, conn_params)
        async with await db.connection() as conn:
            if params:
                if isinstance(params, dict):
                    row = await conn.fetch_one(sql, **params)
                else:
                    row = await conn.fetch_one(sql, *params)
            else:
                row = await conn.fetch_one(sql)

        elapsed_ms = (time.monotonic() - start) * 1000
        row_dict = dict(row) if row and not isinstance(row, dict) else row

        return RowResult(
            driver=self.driver,
            row=row_dict,
            found=row_dict is not None,
            execution_time_ms=round(elapsed_ms, 3),
        )
